# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging
from mocker import ANY

import storm

from offspring.master.master import LexbuilderMaster
from offspring.master.models import (
    Lexbuilder, BuildResult)
from offspring.slave import LexbuilderSlave as Slave
from offspring.master.exceptions import SlaveCommunicationError

from offspring.master.tests.helpers import (
    BaseOffspringMasterTestCase, FakeBuilder,
    StubMailer)
from offspring.config import setup_logging


class TestLexbuilderMaster(BaseOffspringMasterTestCase):

    def setUp(self):
        super(TestLexbuilderMaster, self).setUp()
        self.mailer = StubMailer()

    def get_master(self):
        return LexbuilderMaster(self.config, self.db_store,
                                mailer=self.mailer.mail)

    def create_builder(self, current_state=Slave.STATE_IDLE, arch=None):
        """
        Create and add a Lexbuilder to the store.
        """
        builder = Lexbuilder(u"http://myhost/")
        builder.name = u"test-builder"
        builder.previous_state = Slave.STATE_IDLE
        builder.current_state = current_state
        builder.is_active = True
        builder.is_okay = True
        builder.machine_type = arch or self.project.arch
        self.db_store.add(builder)
        return builder

    def test_instantiate_with_config(self):
        """
        We can instantiate a LexbuilderMaster with a configuration.
        """
        self.config.set("master", "logfile", "%s/offspring-master.log"
                        % self.log_dir)
        setup_logging_mock = self.mocker.replace(setup_logging)
        setup_logging_mock(path=self.config.get("master", "logfile"),
                          level=logging.INFO,
                          fmt="%(asctime)s %(levelname)-8s %(message)s",
                          datefmt="%a, %d %b %Y %H:%M:%S")
        self.mocker.replay()
        master = self.get_master()
        self.assertEqual(self.config, master.config)

    def test_instantiate_master_sets_up_database(self):
        """
        When creating a LexbuilderMaster it should set up logging and
        redirection, and check the database connectivity.
        """
        log_file = self.capture_logging()
        master = self.get_master()
        self.assertTrue(master.is_online)
        log_output = log_file.getvalue()
        self.assertIn("Starting up offspring master", log_output)
        self.assertNotIn("Unable to connect to database.", log_output)

    def test_instantiate_with_missing_database(self):
        """
        If setup is called with a missing database, we should get an
        appropriate error thrown.
        """
        log_file = self.capture_logging()
        broken_uri = "postgres://offspring:temp1234@localhost/no_db"
        self.config.set("master", "db", broken_uri)
        try:
            LexbuilderMaster(self.config)
            raise
        except storm.exceptions.OperationalError:
            log_output = log_file.getvalue()
            self.assertIn("Starting up offspring master", log_output)
            self.assertIn("Unable to connect to database %s." % broken_uri,
                          log_output)

    def test_stop(self):
        """
        Stop marks this builder as not online.
        """
        master = self.get_master()
        self.assertTrue(master.is_online)
        master.stop()
        self.assertFalse(master.is_online)

    def test_scan_slaves_all_builders_idle(self):
        """
        If all Builders are idle, then Lexbuilder.scanSlaves should finish
        after scanning all slaves.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        fake_builder = FakeBuilder(Slave.STATE_IDLE, None)
        master = self.get_master()
        mock_builder = self.mocker.patch(master)
        mock_builder._get_builders()
        self.mocker.result([fake_builder])
        self.mocker.replay()

        master.scanSlaves()
        self.assertTrue(fake_builder.scanned)
        self.assertEqual(
                "Starting up offspring master\n"
                "test-builder: IDLE->IDLE\n"
                "Scan cycle completed\n",
            log_file.getvalue())

    def build_builder_and_slave_with_job_and_result(self):
        """
        This builds and returns a LexbuilderMaster with a single slave builder,
        and a job result.
        """
        builder = self.create_builder()
        builder_mock = self.mocker.patch(builder)
        builder_mock.scan()
        self.mocker.result((Slave.STATE_IDLE, Slave.STATE_IDLE))

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)

        builder_mock.getBuildResult()
        self.mocker.result(build_result)

        builder.current_job_id = build_result.id
        self.mocker.replay()

        master = self.get_master()
        self.subscribe_to_project(self.user, self.project)
        return master, builder

    def test_scan_slaves_with_build_result(self):
        """
        If a Slave has a build result, then we should send a notification.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        master, builder = self.build_builder_and_slave_with_job_and_result()

        master.scanSlaves()

        self.assertTrue(builder.current_job is None)
        self.assertEqual(
            "Starting up offspring master\n"
            "test-builder: IDLE->IDLE\n"
            "test-project has finished building on test-builder\n"
            "Scan cycle completed\n",
            log_file.getvalue())
        self.assertEqual(("test-builder <testing@example.com>",
                          u"testing@example.com"),
                         self.mailer.mails[0][0:2])

    def test_scan_slaves_with_build_result_notification_error(self):
        """
        If when attempting to notify the result of a build, we get an error,
        this error should be logged.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        self.mailer = StubMailer(RuntimeError("Test Error"))
        master, builder = self.build_builder_and_slave_with_job_and_result()

        master.scanSlaves()

        self.assertTrue(builder.current_job is None)
        log_output = log_file.getvalue()
        self.assertIn("Attempt to send build result notification e-mail "
                      "failed.", log_output)
        self.assertIn("Scan cycle completed\n", log_output)
        sender, recpt, message = self.mailer.mails[0]
        self.assertTrue("Subject: Lexbuilder Master (Build Queue Dispatcher)" in
                        message, "Incorrect subject for notification email")

    def test_scan_slaves_with_communication_error(self):
        """
        If we are unable to communicate with a Slave, we should get an
        appropriate error message.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder = self.create_builder()

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)

        builder_mock = self.mocker.patch(builder)

        builder_mock.scan()
        self.mocker.result((Slave.STATE_IDLE, Slave.STATE_IDLE))
        builder_mock.getBuildResult()
        self.mocker.throw(SlaveCommunicationError("Test Error"))
        builder.current_job_id = build_result.id

        master = self.get_master()
        self.subscribe_to_project(self.user, self.project)

        self.mocker.replay()

        master.scanSlaves()

        self.assertEqual(
            "Starting up offspring master\n"
            "%(builder)s: IDLE->IDLE\n"
            "Failed to gather build results from %(builder)s. Test Error\n"
            "Scan cycle completed\n" % {"builder": builder.name},
            log_file.getvalue())

    def test_dispatch_builds(self):
        """
        dispatchBuilds sends BuildRequests to a slave.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder = self.create_builder()

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)

        builder.current_job_id = build_result.id

        builder_mock = self.mocker.patch(builder)
        builder_mock.build(ANY)
        self.mocker.result(builder.current_job)

        master = self.get_master()
        self.mocker.replay()

        master.dispatchBuilds()

        self.assertEqual(
            "Starting up offspring master\n"
            "Project %(name)s needs building (priority: %(priority)d).\n"
            "%(name)s has been dispatched to %(builder)s\n" % {
                "name": self.project.name, "priority": self.project.priority,
                "builder": builder.name},
            log_file.getvalue())

    def test_dispatch_builds_no_builders_available(self):
        """
        If there are no builders available to build a Project, we should log
        this.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder = self.create_builder(current_state=Slave.STATE_BUILDING)

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)

        builder.current_job_id = build_result.id

        master = self.get_master()
        self.mocker.replay()

        master.dispatchBuilds()

        self.assertEqual(
            "Starting up offspring master\n"
            "Project %(name)s needs building (priority: %(priority)d).\n"
            "Unable to find builder to build %(name)s\n" % {
                "name": self.project.name, "priority": self.project.priority},
            log_file.getvalue())

    def test_dispatch_builds_fails_with_communication_error(self):
        """
        If we are unable to communicate with the Slave when attempting to
        request a build, this should be logged.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder = self.create_builder()

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)
        builder.current_job_id = build_result.id

        builder_mock = self.mocker.patch(builder)
        builder_mock.build(ANY)
        self.mocker.throw(SlaveCommunicationError("Test Error"))

        master = self.get_master()
        self.mocker.replay()

        master.dispatchBuilds()

        self.assertEqual(
            "Starting up offspring master\n"
            "Project %(name)s needs building (priority: %(priority)d).\n"
            "Failed to dispatch %(name)s build to %(builder)s. Test Error\n" %
            {"name": self.project.name, "priority": self.project.priority,
             "builder": builder.name},
            log_file.getvalue())

    def test_dispatch_builds_fails_with_no_job(self):
        """
        If the Slave doesn't return a job, then we should log this.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder = self.create_builder()

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)
        builder.current_job_id = build_result.id

        builder_mock = self.mocker.patch(builder)
        builder_mock.build(ANY)
        self.mocker.result(None)

        master = self.get_master()
        self.mocker.replay()

        master.dispatchBuilds()

        self.assertEqual(
            "Starting up offspring master\n"
            "Project %(name)s needs building (priority: %(priority)d).\n"
            "Error when asking %(builder)s to build %(name)s\n" %
            {"name": self.project.name, "priority": self.project.priority,
             "builder": builder.name},
            log_file.getvalue())

    def test_dispatch_to_armhf_builder(self):
        """
        If a project requires an armhf build, we should dispatch to armv7l
        builders.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        builder1 = self.create_builder(arch=u"armv7l")
        builder2 = self.create_builder(arch=u"i386")
        self.project.arch = u"armhf"

        build_request = self.create_build_request()
        build_result = BuildResult(build_request, builder1)
        build_result.name = u"20111115-testing"
        self.db_store.add(build_result)

        builder1.current_job_id = build_result.id

        builder_mock = self.mocker.patch(builder1)
        builder_mock.build(ANY)
        self.mocker.result(builder1.current_job)

        master = self.get_master()
        self.mocker.replay()

        master.dispatchBuilds()

        self.assertEqual(
            "Starting up offspring master\n"
            "Project %(name)s needs building (priority: %(priority)d).\n"
            "%(name)s has been dispatched to %(builder)s\n" % {
                "name": self.project.name, "priority": self.project.priority,
                "builder": builder1.name},
            log_file.getvalue())
