# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import email.parser

from offspring.slave import LexbuilderSlave as Slave

from offspring.master.models import (
    BuildResult, Lexbuilder)

from offspring.master.notifications import (
    notify_crash, notify_problem, notify_buildresult)
from offspring.master.tests.helpers import (
    BaseOffspringMasterTestCase, StubMailer)


class Factory(object):

    def __init__(self):
        self.counter = 1

    def getUniqueString(self):
        self.counter += 1
        return 'str%s' % self.counter


class BaseNotificationTestCase(BaseOffspringMasterTestCase):

    def setUp(self):
        super(BaseNotificationTestCase, self).setUp()
        self.factory = Factory()

    def headerFromMailText(self, msgtext, header):
        msg = email.parser.Parser().parsestr(msgtext)
        return msg[header]


class TestNotifyCrash(BaseNotificationTestCase):

    def test_notify_crash_reply_to_from_config(self):
        """
        notify_crash sets the reply-to of the mail it sends to the value of
        the 'discussion_email' config entry.
        """
        mailer = StubMailer()
        value = self.factory.getUniqueString()
        self.config.set("DEFAULT", "discussion_email", value)
        notify_crash(self.factory.getUniqueString(), self.config, mailer.mail)
        self.assertEqual(
            value, self.headerFromMailText(mailer.mails[0][2], 'reply-to'))

    def test_notify_crash_signs_off_with_instance_name(self):
        """notify_crash signs off with the name of the installation."""
        mailer = StubMailer()
        value = self.factory.getUniqueString()
        self.config.set("DEFAULT", "instance_name", value)
        notify_crash(self.factory.getUniqueString(), self.config, mailer.mail)
        self.assertTrue(mailer.mails[0][2].strip().endswith(value))

    def test_component_name_in_real_name_of_sender(self):
        """
        The 'real name' part of the Sender of the mail sent by notify_crash
        is the 'component' parameter.
        """
        mailer = StubMailer()
        component = self.factory.getUniqueString()
        notify_crash(component, self.config, mailer.mail)
        from_header = mailer.mails[0][0]
        sender_real_name, sender_addr = email.utils.parseaddr(from_header)
        self.assertEqual(component, sender_real_name)

    def test_instance_email_in_address_of_sender(self):
        """
        The 'real name' part of the Sender of the mail sent by notify_crash
        is the 'component' parameter.
        """
        mailer = StubMailer()
        addr = self.factory.getUniqueString()
        self.config.set("DEFAULT", "instance_email", addr)
        notify_crash(self.factory.getUniqueString(), self.config, mailer.mail)
        from_header = mailer.mails[0][0]
        sender_real_name, sender_addr = email.utils.parseaddr(from_header)
        self.assertEqual(addr, sender_addr)


class TestNotifyProblem(BaseNotificationTestCase):

    def test_notify_problem_signs_off_with_instance_name(self):
        """notify_problem signs off with the name of the installation."""
        mailer = StubMailer()
        value = self.factory.getUniqueString()
        self.config.set("DEFAULT", "instance_name", value)
        notify_problem(
            self.factory.getUniqueString(), self.factory.getUniqueString(),
            self.config, mailer.mail)
        self.assertTrue(mailer.mails[0][2].strip().endswith(value))

    def test_component_name_in_real_name_of_sender(self):
        """
        The 'real name' part of the Sender of the mail sent by
        notify_problem is the 'component' parameter.
        """
        mailer = StubMailer()
        component = self.factory.getUniqueString()
        notify_problem(component, self.factory.getUniqueString(), self.config,
                       mailer.mail)
        from_header = mailer.mails[0][0]
        sender_real_name, sender_addr = email.utils.parseaddr(from_header)
        self.assertEqual(component, sender_real_name)

    def test_instance_email_in_address_of_sender(self):
        """
        The 'real name' part of the Sender of the mail sent by
        notify_problem is the 'component' parameter.
        """
        mailer = StubMailer()
        addr = self.factory.getUniqueString()
        self.config.set("DEFAULT", "instance_email", addr)
        notify_problem(
            self.factory.getUniqueString(), self.factory.getUniqueString(),
            self.config, mailer.mail)
        from_header = mailer.mails[0][0]
        sender_real_name, sender_addr = email.utils.parseaddr(from_header)
        self.assertEqual(addr, sender_addr)


class TestNotifyBuildResult(BaseNotificationTestCase):

    def setUp(self):
        super(TestNotifyBuildResult, self).setUp()
        self.builder = Lexbuilder(u"http://myhost/")
        self.builder.name = u"testing"
        self.builder.current_state = Slave.STATE_IDLE
        self.db_store.add(self.builder)
        self.mailer = StubMailer()
        self.build_request = self.create_build_request()
        self.build_result = BuildResult(self.build_request, self.builder)
        self.build_result.name = u"20111115-testing"
        self.db_store.add(self.build_result)

    def test_notify_buildresult(self):
        """
        Send an email users subscribed to a Project with details of the
        build result.
        """
        self.subscribe_to_project(self.user, self.project)
        notify_buildresult(self.build_result, self.config, self.mailer.mail)
        self.assertEqual(self.user.email, self.mailer.mails[0][1])

    def test_notify_buildresult_no_subscriptions(self):
        """
        notify_buildresult should not error in the event that there are no
        subscriptions.
        """
        notify_buildresult(self.build_result, self.config, self.mailer.mail)
        self.assertEqual([], self.mailer.mails)
