# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime

from storm import tz

from offspring.master.utils import (
    build_active_projects, process_pending_build_orders)
from offspring.master.models import (
    BuildRequest, Project, DailyBuildOrder, DailyBuildOrderProject)

from offspring.master.tests.helpers import BaseOffspringMasterTestCase


class TestBuildActiveProjects(BaseOffspringMasterTestCase):

    def test_build_active_projects(self):
        """
        build_active_projects should initiate a BuildRequest for every active
        Project.
        """
        self.assertEqual(0, self.db_store.find(BuildRequest).count())
        build_active_projects(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Nightly build", build_request.reason)

    def test_build_active_projects_ignores_inactive_projects(self):
        """
        build_active_projects should ignore any projects that are not marked as
        active.
        """
        project = Project(u"inactive-project")
        project.config_url = u"http://example.com/"
        project.series = u"inactive-series"
        project.suite = u"inactive-suite"
        project.arch = u"i386"
        project.is_active = False
        self.db_store.add(project)

        self.assertEqual(0, self.db_store.find(BuildRequest).count())
        build_active_projects(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Nightly build", build_request.reason)


class TestProcessPendingBuildOrders(BaseOffspringMasterTestCase):

    def test_process_pending_build_orders(self):
        """
        process_pending_build_orders should identify DailyBuildOrders that are
        scheduled at the same hour, and initiate BuildRequests for them.
        """
        current_hour = datetime.now(tz.tzlocal()).hour
        build_order = DailyBuildOrder(self.project.name, current_hour)
        self.db_store.add(build_order)
        build_order_project = DailyBuildOrderProject()
        build_order_project.dailybuildorder_id = build_order.id
        build_order_project.project_id = self.project.name
        self.db_store.add(build_order_project)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())

        process_pending_build_orders(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Daily build", build_request.reason)

    def test_process_pending_build_orders_ignores_orders_different_times(self):
        """
        process_pending_build_orders should ignore DailyBuildOrders that are
        not scheduled at the current time.
        """
        build_hour = datetime.now(tz.tzlocal()).hour + 1
        build_order = DailyBuildOrder(self.project.name, build_hour)
        self.db_store.add(build_order)
        build_order_project = DailyBuildOrderProject()
        build_order_project.dailybuildorder_id = build_order.id
        build_order_project.project_id = self.project.name
        self.db_store.add(build_order_project)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())

        process_pending_build_orders(self.db_store)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())
