# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging
import platform
import re
import tempfile
import time
import xmlrpclib

from offspring.daemon import LexbuilderDaemon, LexbuilderXMLRPCServer
from offspring.enums import ProjectBuildStates
from offspring.networking import BasicAuthXMLRPCRequestHandler
from offspring.slave.taskmonitor import MonitoredProcess, MonitoredThread


__all__ = ("LexbuilderSlave", "LexbuilderSlaveServer", "ProjectBuildProcess",
           "ShutdownThread")


class LexbuilderSlave(LexbuilderDaemon):

    STATE_IDLE, STATE_BUILDING, STATE_UNKNOWN, STATE_UNAVAIL = (
        u"IDLE",
        u"BUILDING",
        u"UNKNOWN",
        u"UNAVAILABLE"
    )

    def __init__(self, addr, config, server=None):
        self.config = config
        logging.basicConfig(level=logging.DEBUG,
                    format="%(asctime)s %(levelname)-8s %(message)s",
                    datefmt="%a, %d %b %Y %H:%M:%S",
                    filename=self.config.get("slave", "logfile"),
                    filemode="w")
        logging.info(
            "Starting up offspring slave server on port %s" % addr[1])
        self.server = server if server else LexbuilderSlaveServer(
            self, addr, self.config)
        self.state = LexbuilderSlave.STATE_IDLE
        self.currentBuild = None
        LexbuilderDaemon.__init__(self)

    def run(self):
        """
        Start the LexbuilderSlaveServer thread.
        """
        self.server.start()
        logging.info("Slave ready to accept builds")

    def stop(self):
        """
        Terminate any current builds, and stop the LexbuilderSlaveServer
        thread.
        """
        self.stopBuild()
        self.server.stop()
        self.server.join()

    def startBuild(self, projectName, configBranchURL):
        """
        Instigate a new build for projectName, using the configuration at
        configBranchURL.
        """
        self.currentBuild = ProjectBuildProcess(
            self, projectName, configBranchURL)
        logging.info("Starting build: %s %s %s" % (
            self.currentBuild.getCommandPath(), projectName, configBranchURL))
        self.currentBuild.start()

    def stopBuild(self):
        """
        If we have a current build, terminate it.
        """
        if self.currentBuild is not None:
            logging.info("Killing active build.")
            if self.currentBuild.process is not None:
                if self.currentBuild.process.poll() is None:
                    self.currentBuild.process.terminate()
                    if self.currentBuild.process.poll() is None:
                        # Sleep before sending kill.
                        time.sleep(int(self.config.get("slave",
                                                       "child_wait_timeout")))
                        if self.currentBuild.process.poll() is None:
                            self.currentBuild.process.kill()
            self.currentBuild.tidyUp()

    def currentBuildIsActive(self):
        """Return True if current build is active, else False."""
        if self.currentBuild is not None:
            return self.currentBuild.state == ProjectBuildProcess.STATE_ACTIVE
        return False


class LexbuilderSlaveServer(LexbuilderXMLRPCServer):

    def __init__(self, slave, addr, config,
                 requestHandler=BasicAuthXMLRPCRequestHandler,
                 logRequests=False, allow_none=True, encoding="UTF-8",
                 request_timeout=60.0):
        self.slave = slave
        self.config = config
        try:
            request_timeout = self.config.getfloat("slave", "request_timeout")
        except:
            pass
        logging.debug("Request timeout set to %.1f seconds" % request_timeout)
        LexbuilderXMLRPCServer.__init__(
            self, addr, requestHandler, logRequests, allow_none, encoding,
            request_timeout)

    def api_startBuild(self, projectName, configBranchURL):
        """
        Start building the project on the LexbuilderSlave.
        """
        if self.slave.currentBuildIsActive():
            return LexbuilderSlaveServer.REQUEST_ERROR
        self.slave.startBuild(projectName, configBranchURL)
        return LexbuilderSlaveServer.REQUEST_OKAY

    def api_stopBuild(self):
        """
        Stop building the current slave project.

        If the platform version is not compatible, return an error state.
        """
        current_build = self.slave.currentBuild
        if not platform.version() < "2.6" and current_build is not None:
            if current_build.state == ProjectBuildProcess.STATE_ACTIVE:
                self.slave.stopBuild()
                return LexbuilderSlaveServer.REQUEST_OKAY
        return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuilderName(self):
        """Returns the underlying machine network name.

        An empty string is returned if the value cannot be determined.
        """
        return platform.node()

    def api_getMachineType(self):
        """Returns the underlying machine type, e.g. 'i386'.

        An empty string is returned if the value cannot be determined.
        """
        return platform.machine()

    def api_getBuilderState(self):
        """
        Return the state of the LexbuilderSlave.
        """
        return self.slave.state

    def api_getBuildProjectName(self):
        """
        If we have a current build, return the project currently being built.

        Returns an error state if there is no current build.
        """
        if self.slave.currentBuild is not None:
            return self.slave.currentBuild.getProjectName()
        else:
            return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuildName(self):
        """
        If we have a current build, return the requested build name.

        Returns an error state if there is no current build.
        """
        if self.slave.currentBuild is not None:
            return self.slave.currentBuild.getBuildName()
        else:
            return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuildResult(self):
        """
        If we have a current build, return the current build state.

        Returns an error state if there is no current build.
        """
        if self.slave.currentBuild is not None:
            return self.slave.currentBuild.getState()
        else:
            return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuildStartedAt(self):
        """
        If we have a current build, return the time the build was started at.

        If there is no build, or if the build start time is unknown, returns an
        error state.
        """
        if self.slave.currentBuild is not None:
            if self.slave.currentBuild.startedAt is not None:
                return xmlrpclib.DateTime(self.slave.currentBuild.startedAt)
        return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuildFinishedAt(self):
        """
        If we have a current build, return the time the build finished at.

        If there is no build, or if the build finish time is unknown, returns an
        error state.
        """
        if self.slave.currentBuild is not None:
            if self.slave.currentBuild.finishedAt is not None:
                return xmlrpclib.DateTime(self.slave.currentBuild.finishedAt)
        return LexbuilderSlaveServer.REQUEST_ERROR

    def api_getBuildErrorCode(self):
        """
        If we have a current build, with an error code, return it.

        If there is no build, or if the build error is unknown, returns an
        error state.
        """
        if self.slave.currentBuild is not None:
            if self.slave.currentBuild.errorCode is not None:
                return self.slave.currentBuild.errorCode
        return LexbuilderSlaveServer.REQUEST_ERROR

    def api_shutdown(self):
        """
        If the current slave state is IDLE, spawn a thread to shutdown the
        LexbuilderSlave after 1 second.

        Returns REQUEST_OKAY if successful.

        If the slave is not currently IDLE, then return an error state.
        """
        if self.slave.state == LexbuilderSlave.STATE_IDLE:
            self.shutdownThread = ShutdownThread(self)
            self.shutdownThread.start()
            return LexbuilderSlaveServer.REQUEST_OKAY
        else:
            return LexbuilderSlaveServer.REQUEST_ERROR


class ProjectBuildProcess(MonitoredProcess):

    def __init__(self, supervisor, projectName, configBranchURL):
        args = [projectName, configBranchURL]
        MonitoredProcess.__init__(self, supervisor, args)
        self.projectName = projectName

    def setup(self):
        self.buildName = None

    def getState(self):
        if self.state == MonitoredProcess.STATE_ACTIVE:
            return ProjectBuildStates.PENDING
        elif self.state == MonitoredProcess.STATE_FAILED:
            return ProjectBuildStates.FAILED
        elif self.state == MonitoredProcess.STATE_SUCCESS:
            return ProjectBuildStates.SUCCESS
        else:
            return ProjectBuildStates.UNKNOWN

    def preExecution(self):
        logging.debug(
            "Setting state of slave to %s", LexbuilderSlave.STATE_BUILDING)
        self.master.state = LexbuilderSlave.STATE_BUILDING

    def postExecution(self):
        logging.debug(
            "Setting state of slave to %s", LexbuilderSlave.STATE_IDLE)
        self.master.state = LexbuilderSlave.STATE_IDLE

    def getCommandPath(self):
        return self.master.config.get("slave", "build_command")

    def getProjectName(self):
        return self.projectName

    def getBuildName(self):
        if self.buildName is not None:
            return self.buildName
        else:
            if self.state != MonitoredProcess.STATE_ACTIVE:
                for l in self.stdoutdata.split("\n"):
                    if l.startswith("START"):
                        self.buildName = l.split(" ")[3].strip()
                        return self.buildName
            else:
                try:
                    lfile = open(self.master.config.get("slave",
                                                        "build_statefile"))
                    projectName, buildName = lfile.read().split(" ", 1)
                except Exception:
                    logging.exception(
                        "Error when trying to determine build name")
                    return LexbuilderSlaveServer.REQUEST_ERROR
                if projectName == self.getProjectName():
                    self.buildName = buildName.strip("\n")
                    return self.buildName
        return LexbuilderSlaveServer.REQUEST_ERROR


class ShutdownThread(MonitoredThread):

    def run(self):
        time.sleep(1)
        self.master.stop()
