# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging
import os
import platform
import re
import subprocess
import xmlrpclib
from datetime import datetime

from mocker import MockerTestCase

from offspring.tests.helpers import CaptureLoggingTestCase
from offspring.slave.tests.helpers import LexbuilderSlaveTestMixin

from offspring.enums import ProjectBuildStates
from offspring.slave import (
    LexbuilderSlave, ProjectBuildProcess, LexbuilderSlaveServer,
    ShutdownThread)


class FakeLexbuilderSlaveServer(object):
    def start(self):
        self.started = True

    def stop(self):
        self.stopped = True

    def join(self):
        self.joined = True


class FakeBuild(object):

    class FakeBuildProcess(object):

        def poll(self):
            return None

        def terminate(self):
            self.terminated = True

        def kill(self):
            self.killed = True

    def __init__(self, project_name=None, name=None, started_at=None,
                 finished_at=None, error_code=None,
                 state=ProjectBuildProcess.STATE_SUCCESS):
        self.process = self.FakeBuildProcess()
        self.state = state
        self.startedAt = started_at
        self.finishedAt = finished_at
        self.project_name = project_name
        self.name = name
        self.errorCode = error_code

    def getState(self):
        return self.state

    def getProjectName(self):
        return self.project_name

    def getBuildName(self):
        return self.name

    def tidyUp(self):
        self.tidied = True


class FakeProjectBuildProcess(object):
    state = ProjectBuildProcess.STATE_ACTIVE


class TestLexbuilderSlave(LexbuilderSlaveTestMixin, CaptureLoggingTestCase):
    """
    Tests for the LexbuilderSlave class.
    """

    def setUp(self):
        super(TestLexbuilderSlave, self).setUp()
        self.fake_server = FakeLexbuilderSlaveServer()
        self.slave = LexbuilderSlave(("localhost", 8760), self.config,
                                     self.fake_server)

    def test_slave_run_starts_server_and_logs(self):
        """
        LexbuilderSlave.run should start the LexbuilderSlaveServer.
        """
        log_file = self.capture_logging()
        self.slave.run()
        self.assertEqual("Slave ready to accept builds\n", log_file.getvalue())
        self.assertTrue(self.fake_server.started)

    def test_slave_stop(self):
        """
        LexbuilderSlave.stop, should terminate any current build, and stop
        the LexbuilderSlaveServer.
        """
        slave_mock = self.mocker.patch(self.slave)
        slave_mock.stopBuild()
        self.mocker.replay()

        self.slave.stop()
        self.assertTrue(self.fake_server.stopped)
        self.assertTrue(self.fake_server.joined)

    def test_slave_stop_build_no_current_build(self):
        """
        LexbuilderSlave.stopBuild should do nothing if there's no current
        build in progress.
        """
        log_file = self.capture_logging()
        self.slave.stopBuild()
        self.assertEqual("", log_file.getvalue())

    def test_slave_stop_build_with_current_build(self):
        """
        LexbuilderSlave.stopBuild should terminate the current build, this
        involves killing the process running the build.
        """
        self.slave.currentBuild = fake_build = FakeBuild()
        log_file = self.capture_logging()

        self.slave.stopBuild()

        self.assertTrue(fake_build.process.terminated)
        self.assertTrue(fake_build.process.killed)
        self.assertTrue(fake_build.tidied)
        self.assertEqual("Killing active build.\n", log_file.getvalue())

    def test_slave_start_build(self):
        """
        LexbuilderSlave.startBuild should start a new build on the server.
        """
        log_file = self.capture_logging()
        build_process_mock = self.mocker.patch(ProjectBuildProcess)
        build_process_mock.start()
        self.mocker.replay()

        fake_url = u"http://example.com/project"
        self.slave.startBuild(u"testing", fake_url)
        self.assertEqual(["testing", fake_url],
                         self.slave.currentBuild.arguments)
        self.assertEqual(
            "Starting build: /usr/bin/offspring-build testing %s\n" % fake_url,
            log_file.getvalue())

    def test_current_build_is_active_no_current_build(self):
        """Test currentBuildIsActive returns False when currentBuild = False"""

        # To start with, currentBuild should be None, so active == False
        self.assertEqual(self.slave.currentBuild, None)
        self.assertEqual(self.slave.currentBuildIsActive(), False)

    def test_current_build_is_active_build_in_active_state(self):
        """Test currentBuildIsActive returns True when build is active"""
        # Set up a fake build process (starts off in STATE_ACTIVE)
        self.slave.currentBuild = FakeProjectBuildProcess()

        # Now we have started a build, currentBuild has been assigned.
        self.assertNotEqual(self.slave.currentBuild, None)
        self.assertEqual(self.slave.currentBuild.state,
                         ProjectBuildProcess.STATE_ACTIVE)
        self.assertEqual(self.slave.currentBuildIsActive(), True)

    def test_current_build_is_active_build_state_not_active(self):
        """Test currentBuildIsActive when currentBuild state != STATE_ACTIVE

        currentBuildIsActive should return False when
        slave.currentBuild.state != STATE_ACTIVE. Clearly, currentBuild needs
        to be valid for this test.

        """
        # Set up a fake build process (starts off in STATE_ACTIVE)
        self.slave.currentBuild = FakeProjectBuildProcess()

        self.slave.currentBuild.state = ProjectBuildProcess.STATE_SUCCESS
        self.assertEqual(self.slave.currentBuildIsActive(), False)

        self.slave.currentBuild.state = ProjectBuildProcess.STATE_FAILED
        self.assertEqual(self.slave.currentBuildIsActive(), False)


class FakeLexbuilderSlave(object):
    """Fake slave for use in tests."""
    def __init__(self, config=None, state=LexbuilderSlave.STATE_IDLE):
        self.project_name = None
        self.config_url = None
        self.stopped = False
        self.state = state
        self.config = config

    def startBuild(self, project_name, config_url):
        self.project_name = project_name
        self.config_url = config_url
        self.started = True

    def stopBuild(self):
        self.stopped = True

    def currentBuildIsActive(self):
        """Return True if current build is active, else False."""
        if self.currentBuild is not None:
            return self.currentBuild.state == ProjectBuildProcess.STATE_ACTIVE
        return False


class TestLexbuilderSlaveServer(LexbuilderSlaveTestMixin,
                                CaptureLoggingTestCase):
    """
    Tests for the LexbuilderSlaveServer class.
    """

    def setUp(self):
        super(TestLexbuilderSlaveServer, self).setUp()
        self.fake_slave = FakeLexbuilderSlave(self.config)

    def create_slave_server(self):
        """
        Instantiate a slave server for the fake_slave.
        """
        return LexbuilderSlaveServer(
            self.fake_slave, ("localhost", 8760), self.config)

    def test_create(self):
        """
        We can instantiate a new LexbuilderSlaveServer.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)

        slave_server = self.create_slave_server()
        self.assertEqual(self.fake_slave, slave_server.slave)
        self.assertEqual(("127.0.0.1", 8760), slave_server.server_address)
        self.assertEqual(self.config, slave_server.config)
        self.assertEqual("Request timeout set to 60.0 seconds\n",
                         log_file.getvalue())

    def test_create_with_missing_request_timeout(self):
        """
        LexbuilderSlaveServer should set the request timeout to the value
        specified in the configuration.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        self.config.set("slave", "request_timeout", "30.0")

        slave_server = LexbuilderSlaveServer(
            self.fake_slave, ("localhost", 8760), self.config)
        self.assertEqual("Request timeout set to 30.0 seconds\n",
                         log_file.getvalue())

    def test_create_with_invalid_configuration(self):
        """
        LexbuilderSlaveServer should handle invalid request timeout values by
        falling back to the default.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)
        self.config.set("slave", "request_timeout", "broken value")


        slave_server = LexbuilderSlaveServer(
            self.fake_slave, ("localhost", 8760), self.config)
        self.assertEqual("Request timeout set to 60.0 seconds\n",
                         log_file.getvalue())

    def test_api_start_build(self):
        """
        If the server doesn't have an active build, then api_startBuild
        should start a build on the slave with the correct project name
        and configuration URL.
        """
        self.fake_slave.currentBuild = FakeBuild()
        slave_server = self.create_slave_server()

        response = slave_server.api_startBuild("testing",
                                               "http://example.com/")
        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY, response)
        self.assertEqual("testing", self.fake_slave.project_name)
        self.assertEqual("http://example.com/", self.fake_slave.config_url)
        self.assertTrue(self.fake_slave.started)

    def test_api_start_build_with_busy_server(self):
        """
        If the server has an active build in progress, api_startBuild should
        return an error state.
        """
        slave_server = self.create_slave_server()

        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_startBuild("testing",
                                               "http://example.com/")
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.project_name)
        self.assertFalse(self.fake_slave.config_url)

    def test_api_stop_build(self):
        """
        If the SlaveServer has an active build, api_stopBuild should end the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.stopped)

    def test_api_stop_build_with_correct_platform_version(self):
        """
        If platform.version is > 2.6, and we have a current build, then we
        can stop the build slave.

        FIXME: Add documentation for why?
        """
        version_mock = self.mocker.replace("platform.version")
        version_mock()
        self.mocker.result("2.7")
        self.mocker.replay()
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY, response)
        self.assertTrue(self.fake_slave.stopped)

    def test_api_stop_build_no_current_build(self):
        """
        If the SlaveServer does not have an active build, api_stopBuild should
        return an error state.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.stopped)

    def test_api_get_builder_name(self):
        """
        LexbuilderSlaveServer.api_getBuilderName should return the network name
        for this machine.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(platform.node(), slave_server.api_getBuilderName())

    def test_api_get_machine_type(self):
        """
        LexbuilderSlaveServer.api_getMachineType should return the platform the
        builder is running on.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(platform.machine(), slave_server.api_getMachineType())

    def test_api_get_builder_state(self):
        """
        LexbuilderSlaveServer.api_getBuilderState should return the current
        state of the underlying LexBuilderSlave.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(LexbuilderSlave.STATE_IDLE,
                         slave_server.api_getBuilderState())

    def test_api_get_build_project_name(self):
        """
        LexbuilderSlaveServer.api_getBuildProjectName should return the name of
        the current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(project_name=u"testproject")
        self.assertEqual(u"testproject",
                         slave_server.api_getBuildProjectName())

    def test_api_get_build_project_name_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildProjectName should return an error
        state if there's no current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildProjectName())

    def test_api_get_build_name(self):
        """
        LexbuilderSlaveServer.getBuildName should return the name of the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(name=u"testbuild")
        self.assertEqual(u"testbuild",
                         slave_server.api_getBuildName())

    def test_api_get_build_name_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildName should return an error
        state if there's no current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildName())

    def test_api_get_build_result(self):
        """
        LexbuilderSlaveServer.getBuildResult should return the result of the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(ProjectBuildProcess.STATE_SUCCESS,
                         slave_server.api_getBuildResult())

    def test_api_get_build_result_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildResult should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildResult())

    def test_api_get_build_started_at(self):
        """
        LexbuilderSlaveServer.getBuildStartedAt should return the time the
        current build (if any) started at, using the xmlrpc DateTime
        serialiser.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(started_at=datetime.today())
        self.fake_slave.currentBuild = current_build
        self.assertEqual(xmlrpclib.DateTime(current_build.startedAt),
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_started_at_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildStartedAt should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_started_at_with_no_start_time(self):
        """
        LexbuilderSlaveServer.api_getBuildStartedAt should return an error if
        the start time of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_finished_at(self):
        """
        LexbuilderSlaveServer.getBuildFinishedAt should return the time the
        current build (if any) finished at, using the xmlrpc DateTime
        serialiser.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(finished_at=datetime.today())
        self.fake_slave.currentBuild = current_build
        self.assertEqual(xmlrpclib.DateTime(current_build.finishedAt),
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_finished_at_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildFinishedAt should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_finished_at_with_no_finish_time(self):
        """
        LexbuilderSlaveServer.api_getBuildFinishedAt should return an error if
        the finish time of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_error_code(self):
        """
        LexbuilderSlaveServer.getBuildErrorCode should return the current error
        code of the build.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(error_code=1)
        self.fake_slave.currentBuild = current_build
        self.assertEqual(1, slave_server.api_getBuildErrorCode())

    def test_api_get_build_error_code_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildErrorCode should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildErrorCode())

    def test_api_get_build_error_code_with_no_finish_time(self):
        """
        LexbuilderSlaveServer.api_getBuildErrorCode should return an error if
        the error code of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildErrorCode())

    def test_api_shutdown_slave_not_idle(self):
        """
        LexbuilderSlaveServer.api_shutdown should return an error if the slave
        is not idle.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.fake_slave.state = ProjectBuildProcess.STATE_ACTIVE
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_shutdown())

    def test_api_shutdown_slave_idle(self):
        """
        LexbuilderSlaveServer.api_shutdown should start a shutdown thread if
        the slave is currently idle.
        """
        class FakeShutdownThread(object):
            def start(self):
                self.started = True
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()

        mock_shutdown_thread = self.mocker.replace(ShutdownThread,
                                                   passthrough=False)
        mock_shutdown_thread(slave_server)
        fake_shutdown = FakeShutdownThread()
        self.mocker.result(fake_shutdown)
        self.mocker.replay()

        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY,
                         slave_server.api_shutdown())
        self.assertTrue(fake_shutdown.started)


class TestProjectBuildProcess(LexbuilderSlaveTestMixin, CaptureLoggingTestCase):

    def setUp(self):
        super(TestProjectBuildProcess, self).setUp()
        self.master = FakeLexbuilderSlave(config=self.config)
        self.build_process = ProjectBuildProcess(
            self.master, u"testing", "http://example.com/project")

    def test_create(self):
        """
        We can instantiate a new ProjectBuildProcess.
        """
        self.assertEqual(u"testing", self.build_process.projectName)
        self.assertEqual([u"testing", "http://example.com/project"],
                         self.build_process.arguments)

    def test_get_state_currently_active(self):
        """
        If ProjectBuildProcess.state is STATE_ACTIVE then
        ProjectBuildProcess.get_state() should return
        ProjectBuildStates.PENDING.
        """
        self.build_process.state = self.build_process.STATE_ACTIVE
        self.assertEqual(ProjectBuildStates.PENDING,
                         self.build_process.getState())

    def test_get_state_currently_failed(self):
        """
        If ProjectBuildProcess.state is STATE_FAILED then
        ProjectBuildProcess.get_state() should return
        ProjectBuildStates.FAILED.
        """
        self.build_process.state = self.build_process.STATE_FAILED
        self.assertEqual(ProjectBuildStates.FAILED,
                         self.build_process.getState())

    def test_get_state_currently_success(self):
        """
        If ProjectBuildProcess.state is STATE_SUCCESS then
        ProjectBuildProcess.get_state() should return
        ProjectBuildStates.SUCCESS.
        """
        self.build_process.state = self.build_process.STATE_SUCCESS
        self.assertEqual(ProjectBuildStates.SUCCESS,
                         self.build_process.getState())

    def test_get_state_currently_unknown_state(self):
        """
        If ProjectBuildProcess.state is not one of STATE_ACTIVE, STATE_FAILED
        or STATE_SUCCESS ProjectBuildProcess.get_state() should return
        ProjectBuildStates.UNKNOWN.
        """
        self.build_process.state = None
        self.assertEqual(ProjectBuildStates.UNKNOWN,
                         self.build_process.getState())

    def test_pre_execution(self):
        """
        ProjectBuildProcess.preExecution should log the fact that the slave is
        starting a build, and set the current state of the LexbuilderSlave to
        STATE_BUILDING.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)

        self.build_process.preExecution()
        self.assertEqual(LexbuilderSlave.STATE_BUILDING,
                         self.build_process.master.state)
        self.assertEqual("Setting state of slave to BUILDING\n",
                         log_file.getvalue())

    def test_post_execution(self):
        """
        ProjectBuildProcess.preExecution should log the fact that the slave is
        finishing a build, and set the current state of the LexbuilderSlave to
        STATE_IDLE.
        """
        log_file = self.capture_logging(log_level=logging.DEBUG)

        self.build_process.postExecution()
        self.assertEqual(LexbuilderSlave.STATE_IDLE,
                         self.build_process.master.state)
        self.assertEqual("Setting state of slave to IDLE\n",
                         log_file.getvalue())

    def test_get_command_path(self):
        """
        ProjectBuildProcess.getCommandPath should return the configured
        build_command.
        """
        self.config.set("slave", "build_command",
                        "/usr/bin/offspring-build")
        self.assertEqual("/usr/bin/offspring-build",
                         self.build_process.getCommandPath())

    def test_get_project_name(self):
        """
        ProjectBuildProcess.getProjectName should return the name of the
        project currently configured for building.
        """
        self.build_process.projectName = u"testing"
        self.assertEqual(u"testing",
                         self.build_process.getProjectName())

    def test_get_build_name_with_known_build_name(self):
        """
        ProjectBuildProcess.getBuildName should return the name of the current
        build if known.
        """
        self.build_process.buildName = u"testing"
        self.assertEqual(u"testing", self.build_process.getBuildName())

    def test_get_build_name_with_unknown_build_inactive_process(self):
        """
        If the build name is unknown, and the process is not currently active,
        then we ProjectBuildProcess.getBuildName should examine the output of
        the build process and look for START <projectname> <buildname> and
        extract the buildname and return it.
        """
        self.build_process.stdoutdata = "START project testproject mybuild\n"
        self.assertEqual("mybuild", self.build_process.getBuildName())

    def test_get_build_name_with_unknown_build_active_process(self):
        """
        If the build name is unknown, and the process is currently active,
        then we ProjectBuildProcess.getBuildName should examine the the build
        statefile from the configuration and parse the project name and build
        name from there.
        """
        self.build_process.state = self.build_process.STATE_ACTIVE
        self.build_process.projectName = "newproject"
        build_statefile_name = self.makeFile("newproject newbuild\n")
        self.config.set("slave", "build_statefile", build_statefile_name)
        self.assertEqual("newbuild", self.build_process.getBuildName())

    def test_get_build_name_with_unknown_build_active_invalid_project(self):
        """
        If the build name is unknown, and the process is currently active,
        then we ProjectBuildProcess.getBuildName should examine the the build
        statefile from the configuration and parse the project name and build
        name from there.

        If the project name parsed from the build file doesn't match the
        project we're building, then we should get an error.
        """
        self.build_process.state = self.build_process.STATE_ACTIVE
        self.build_process.projectName = "anotherprojectentirely"
        build_statefile_name = self.makeFile("newproject newbuild\n")
        self.config.set("slave", "build_statefile", build_statefile_name)

        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         self.build_process.getBuildName())

    def test_get_build_name_with_unknown_build_active_process_error(self):
        """
        If the build name is unknown, and the process is currently active,
        then we ProjectBuildProcess.getBuildName should examine the the build
        statefile from the configuration, any error reading this should result
        in an error return.
        """
        self.build_process.state = self.build_process.STATE_ACTIVE
        self.build_process.projectName = "newproject"
        build_statefile_name = self.makeFile()
        self.config.set("slave", "build_statefile", build_statefile_name)
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         self.build_process.getBuildName())


class TestShutdownThread(MockerTestCase):
    """
    Tests for the ShutdownThread class.
    """

    def test_run(self):
        """
        """
        sleep_mock = self.mocker.replace("time.sleep")
        sleep_mock(1)
        self.mocker.replay()
        master = FakeLexbuilderSlaveServer()

        shutdown_thread = ShutdownThread(master)
        shutdown_thread.run()

        self.assertTrue(master.stopped)
