# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import sys
import logging
from cStringIO import StringIO

from mocker import MockerTestCase

from offspring.config import (
    get_configuration, setup_logging, print_builder_config)


class CaptureLoggingTestCase(MockerTestCase):

    def setUp(self):
        super(CaptureLoggingTestCase, self).setUp()
        self.loggers = []

    def tearDown(self):
        for logger, handler, old_level in reversed(self.loggers):
            logger.removeHandler(handler)
            logger.setLevel(old_level)
        super(CaptureLoggingTestCase, self).tearDown()

    def capture_logging(self, logger_name="", log_level=logging.INFO,
                        format=""):
        """
        Call this with an optional logger_name and this will log the named
        logger into a log_file (StringIO) that is returned.
        """
        log_file = StringIO()
        log_handler = logging.StreamHandler(log_file)
        logger = logging.getLogger(logger_name)
        logger.addHandler(log_handler)

        if format:
            log_handler.setFormatter(logging.Formatter(format))

        old_logger_level = logger.level
        if log_level:
            logger.setLevel(log_level)
        self.loggers.append((logger, log_handler, old_logger_level))
        return log_file


class TestOffspringConfiguration(CaptureLoggingTestCase):

    def setUp(self):
        super(TestOffspringConfiguration, self).setUp()
        self.fake_environment = {}
        self.addCleanup(setattr, os, "getenv", os.getenv)
        os.getenv = self.fake_get_environment
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = self.fake_environment
        self.expected_base_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', '..', '..'))

    def fake_get_environment(self, key):
        return self.fake_environment.get(key)

    def test_explicit_path(self):
        """
        If we pass an explicit path to get_configuration, it should be used to
        read the file and fail if it doesn't exist.
        """
        config_filename = self.makeFile("[DEFAULT]\nbase_dir=/srv/offspring\n"
                                        "[test]\nvalue=1\n")
        configuration = get_configuration(config_filename)

        self.assertEqual(1, configuration.getint("test", "value"))
        self.assertEqual(config_filename,
                         configuration.get("DEFAULT", "config_filename"))
        self.assertEqual("/srv/offspring",
                         configuration.get("DEFAULT", "base_dir"))

    def test_supplement_config_with_hostname_and_username(self):
        """
        The Configuration parsed from the file should be supplemented with
        additional variables.

        The user running Offspring and the Hostname it's run on should be
        inserted into the DEFAULT section.
        """
        filename = self.makeFile("[test]\nvalue=1\n")
        mock_uname = self.mocker.replace(os.uname)
        mock_uname()
        self.mocker.result(("Linux", "aurore", "3.0.0-13-generic",
                            "#22-Ubuntu SMP Wed Nov 2 13:27:26 UTC 2011",
                            "x86_64", "86_64"))
        self.mocker.replay()
        os.environ["USERNAME"] = "testing"

        configuration = get_configuration(filename)

        self.assertEqual(1, configuration.getint("test", "value"))
        self.assertEqual("aurore", configuration.get("DEFAULT", "hostname"))
        self.assertEqual("testing", configuration.get("DEFAULT", "username"))

    def test_default_filename(self):
        """
        By default, get_configuration should read
        <base_dir>config/offspring.cfg if no other configuration
        file is specified.
        """
        config_filename = os.path.join(self.expected_base_dir, "config",
                                       "offspring.cfg")
        path_exists_mock = self.mocker.replace("os.path.exists")
        path_exists_mock(config_filename)
        self.mocker.result(True)
        open_mock = self.mocker.replace("__builtin__.open")
        open_mock(config_filename, "r")
        self.mocker.result(StringIO("[test]\nvalue=3\n"))
        self.mocker.replay()

        configuration = get_configuration()

        self.assertEqual(3, configuration.getint("test", "value"))
        self.assertEqual(config_filename,
                         configuration.get("DEFAULT", "config_filename"))

    def test_supplement_config_with_basedir(self):
        """
        The Configuration parsed from the file should be supplemented with
        additional variables.

        The base_dir variable should reflect the path to the base 'offspring'
        directory, which is above here.
        """
        config_filename = os.path.join(self.expected_base_dir, "config",
                                       "offspring.cfg")
        path_exists_mock = self.mocker.replace("os.path.exists")
        path_exists_mock(config_filename)
        self.mocker.result(True)
        open_mock = self.mocker.replace("__builtin__.open")
        open_mock(config_filename, "r")
        self.mocker.result(StringIO("[test]\nvalue=3\n"))
        self.mocker.replay()

        configuration = get_configuration()

        self.assertEqual(self.expected_base_dir,
                         configuration.get("DEFAULT", "base_dir"))

    def test_set_environment(self):
        """
        If no file is passed to get_configuration, then we should attempt to
        find a path in the environment.
        """
        config_filename = self.makeFile("[DEFAULT]\nbase_dir=/srv/offspring\n"
                                        "[test]\nvalue=2\n")
        os.environ["OFFSPRING_CONFIGFILE"] = config_filename

        configuration = get_configuration()

        self.assertEqual(2, configuration.getint("test", "value"))
        self.assertEqual("/srv/offspring",
                         configuration.get("DEFAULT", "base_dir"))

    def test_print_builder_config(self):
        """
        print_builder_config should write the elements from the configuration in
        the "builder" and "archiver" sections to stdout.
        """
        config_filename = self.makeFile("[builder]\n"
                                        "checksum_types: sha1 md5\n"
                                        "sign_builds: enabled\n"
                                        "[archiver]\n"
                                        "archive_mode: 0\n"
                                        "archive_threshold: 7\n")
        configuration = get_configuration(config_filename)
        stdout_mock = self.mocker.replace(sys.stdout)
        stdout_mock.write('IBS_CHECKSUM_TYPES="sha1 md5"\n')
        stdout_mock.write('IBS_SIGN_BUILDS="enabled"\n')
        stdout_mock.write('IBS_ARCHIVE_MODE="0"\n')
        stdout_mock.write('IBS_ARCHIVE_THRESHOLD="7"\n')

        self.mocker.replay()
        print_builder_config(configuration)

    def test_print_builder_config_ignores_sections(self):
        """
        print_builder_config should not write other sections to stdout.
        """
        config_filename = self.makeFile("[server]\n"
                                        "testing: 1\n"
                                        "value: 10\n"
                                        "[builder]\n"
                                        "[archiver]\n")
        configuration = get_configuration(config_filename)
        stdout_mock = self.mocker.replace(sys.stdout)
        self.expect(stdout_mock.write('IBS_TESTING="1"\n')).count(0)
        self.expect(stdout_mock.write('IBS_VALUE="10"\n')).count(0)
        self.mocker.replay()

        print_builder_config(configuration)

    def test_print_builder_config_ignores_duplicate_values(self):
        """
        print_builder_config should ignore duplicate key names in the [builder]
        and [archiver] sections.
        """
        config_filename = self.makeFile("[builder]\n"
                                        "checksum_types: sha1 md5\n"
                                        "sign_builds: enabled\n"
                                        "[archiver]\n"
                                        "sign_builds: enabled\n"
                                        "archive_mode: 0\n"
                                        "archive_threshold: 7\n")
        configuration = get_configuration(config_filename)
        stdout_mock = self.mocker.replace(sys.stdout)
        stdout_mock.write('IBS_CHECKSUM_TYPES="sha1 md5"\n')
        self.expect(stdout_mock.write('IBS_SIGN_BUILDS="enabled"\n')).count(1)
        stdout_mock.write('IBS_ARCHIVE_MODE="0"\n')
        stdout_mock.write('IBS_ARCHIVE_THRESHOLD="7"\n')
        self.mocker.replay()

        print_builder_config(configuration)

    def test_print_builder_config_defaults_to_configuration(self):
        """
        print_builder_config should fallback to get_configuration() if no
        configuration is supplied.
        """
        config_filename = self.makeFile("[builder]\n"
                                        "checksum_types: sha1 md5\n"
                                        "sign_builds: enabled\n"
                                        "[archiver]\n"
                                        "sign_builds: enabled\n"
                                        "archive_mode: 0\n"
                                        "archive_threshold: 7\n")
        configuration = get_configuration(config_filename)

        get_configuration_mock = self.mocker.replace(get_configuration)
        get_configuration_mock()
        self.mocker.result(configuration)

        stdout_mock = self.mocker.replace(sys.stdout)
        stdout_mock.write('IBS_CHECKSUM_TYPES="sha1 md5"\n')
        self.expect(stdout_mock.write('IBS_SIGN_BUILDS="enabled"\n')).count(1)
        stdout_mock.write('IBS_ARCHIVE_MODE="0"\n')
        stdout_mock.write('IBS_ARCHIVE_THRESHOLD="7"\n')
        self.mocker.replay()

        print_builder_config()


class SetupLoggingTest(MockerTestCase):

    def test_setup_logging_with_path(self):
        """
        L{setup_logging} takes a name and a path and configures a logger
        with the specified name to write to the stream.
        """
        path = self.makeFile()
        handler = setup_logging(path=path)
        log = logging.getLogger()
        log.info("This is a test!")
        handler.close()

        stream = open(path, "r")
        self.assertIn("This is a test!", stream.read())

    def test_setup_logging_with_path_works_with_log_rotation(self):
        """
        The special C{WatchedFileHandler} is used when a path is provided,
        ensuring correct behaviour when log rotate is used.
        """
        path = self.makeFile()
        handler = setup_logging(path=path)
        log = logging.getLogger()
        log.info("This is a test!")

        os.unlink(path)
        log.info("This a message logged after rotation.")
        handler.close()

        stream = open(path, "r")
        self.assertIn("This a message logged after rotation.", stream.read())
