# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.forms import (
    ModelChoiceField,
    ModelMultipleChoiceField,
    ModelForm,
    Textarea,
    TextInput,
)
from django.forms import fields

from django_group_access.middleware import get_access_control_user
from django_group_access.models import AccessGroup

from offspring.web.queuemanager.models import (
    Project,
    ProjectGroup,
    Release,
    LaunchpadProject
)
from offspring.web.queuemanager.widgets import SelectWithAddNew


class ProjectBaseForm(ModelForm):
    status = fields.CharField(max_length=200,
        widget=fields.Select(choices=Project.STATUS_CHOICES), required=True)
    access_groups = ModelMultipleChoiceField(
        queryset=AccessGroup.objects.all(), required=False,
        help_text='Restrict access to these groups. Selecting no groups here '
        'will make this a public project.')

    def __init__(self, *args, **kwargs):
        super(ProjectBaseForm, self).__init__(*args, **kwargs)
        logged_in_user = get_access_control_user()
        if logged_in_user and not logged_in_user.is_superuser:
            self.fields['access_groups'].queryset = \
                self.fields['access_groups'].queryset.filter(
                    members=logged_in_user)

    class Meta:
        model = Project
        # N.B. We don't override the widget type for 'project_group and 
        # 'launchpad_project' here as this form is used by django admin and
        # we want it to continue to use it's own widget for those fields.
        widgets = {
            'name' : TextInput(attrs={'style': 'text-transform: lowercase;'}),
            'series' : TextInput(attrs={'style': 'text-transform: lowercase;'}),
            'config_url': TextInput(attrs={'size': 50}),
            'notes' : Textarea(attrs={'cols': 73, 'rows' : 4}),
        }
        exclude = ('owner',)

    def clean_name(self):
        return self.cleaned_data['name'].lower()

    def clean_series(self):
        return self.cleaned_data['series'].lower()


class CreateProjectForm(ProjectBaseForm):
    class Meta(ProjectBaseForm.Meta):
        exclude = ('priority', 'is_active', 'suite', 'owner')
        widgets = {
            'project_group' : SelectWithAddNew(),
            'launchpad_project' : SelectWithAddNew(),
        }
        widgets.update(ProjectBaseForm.Meta.widgets)


class EditProjectForm(ProjectBaseForm):
    class Meta(ProjectBaseForm.Meta):
        exclude = ('name', 'priority', 'is_active', 'owner')
        widgets = {
            'project_group' : SelectWithAddNew(),
            'launchpad_project' : SelectWithAddNew(),
        }
        widgets.update(ProjectBaseForm.Meta.widgets)


class ReleaseForm(ModelForm):
    class Meta:
        model = Release
        fields = ( 'name', 'milestone', 'tag', 'checklist_url', 'notes')


class LaunchpadProjectForm(ModelForm):
    class Meta:
        model = LaunchpadProject
        widgets = {
            'name' : TextInput(attrs={'style': 'text-transform: lowercase;'}),
        }

    def clean_name(self):
        return self.cleaned_data['name'].lower()


class ProjectGroupForm(ModelForm):
    class Meta:
        model = ProjectGroup
        widgets = {
            'name' : TextInput(attrs={'style': 'text-transform: lowercase;'}),
        }

    def clean_name(self):
        return self.cleaned_data['name'].lower()
