# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from django.db import models

from offspring.enums import ProjectBuildStates


def delta_to_seconds(delta):
    """
    Return the total amount of time represented by a timedelta.
    """
    return ((delta.microseconds + (delta.seconds + delta.days
                                   * 24 * 3600)
            * 10 ** 6) / 10 ** 6)


def get_average_for_timeperiods(queryset, columns):
    """
    Return the average time between two columns in a QuerySet, or None if there
    are no matching results.

    queryset: A QuerySet of BuildResults to be averaged.
    columns: A sequence of two column names to be extracted, this must be
        start_time, end_time.
    """
    time_stamps = queryset.values_list(*columns)
    if not time_stamps:
        return None

    times = [delta_to_seconds(end_time - start_time)
        for (start_time, end_time) in time_stamps]
    return float(sum(times)) / len(time_stamps)


class BuildResultMetricsManager(models.Manager):

    def get_average_build_time(self, project=None, automated=None, period=None,
                               project_group=None):
        """
        Return the average time taken between BuildResult.started_at and
        BuildResult.finished_at in seconds for all BuildResults associated with
        the provided Project.

        project: a Project, if None is supplied, averages across all models.
        period: A tuple of datetime objects (start_time, end_time)
        automated: If True, then only select BuildResults that were not
            initiated by a user, if False, then only select BuildResults that
            were initiated by a user, otherwise, select all BuildRequests.
        project_group: If provided, this will compute the average build time
            for a ProjectGroup.

        NOTE: If a project is provided, this will override the project_group
            argument.
        """
        query_params = {"result__exact": ProjectBuildStates.SUCCESS}
        if period is not None:
            query_params["finished_at__range"] = period
        else:
            query_params["finished_at__isnull"] = False

        return self._get_average_time_for_query(
            query_params, project, project_group, automated,
            ("started_at", "finished_at"))

    def _get_average_time_for_query(
        self, query_params, project, project_group, automated, columns):
        """
        Calculate the average time for a BuildResult query.

        query_params: dictionary of ORM query parameters.
        project: Project to restrict queries.
        project_group: ProjectGroup to restrict queries.
        automated: Whether or not automated builds should be queried.
        columns: The columns from the resulting QuerySet that should be
            averaged.
        """
        if project is not None:
            query_params["project"] = project
        else:
            if project_group is not None:
                query_params["project__project_group"] = project_group

        if automated is not None:
            query_params["requestor__isnull"] = automated

        build_results = self.model.objects.filter(**query_params)
        return get_average_for_timeperiods(build_results, columns)

    def get_average_queue_time(self, project=None, automated=None, period=None,
                               project_group=None):
        """
        Return the average time taken between BuildResult.requested_at and
        BuildResult.dispatched_at in seconds for all BuildResults associated
        with the provided Project.

        project: a Project, if None is supplied, averages across all models.
        period: A tuple of datetime objects (start_time, end_time)
        automated: If True, then only select BuildResults that were not
            initiated by a user, if False, then only select BuildResults that
            were initiated by a user, otherwise, select all BuildRequests.
        project_group: If provided, this will compute the average queuetime
            for a ProjectGroup.
        NOTE: If a project is provided, this will override the project_group
            argument.
        """
        query_params = {}
        if period is not None:
            query_params["dispatched_at__range"] = period
        else:
            query_params["dispatched_at__isnull"] = False

        return self._get_average_time_for_query(
            query_params, project, project_group, automated,
            ("requested_at", "dispatched_at"))
