from itertools import count

from django.contrib.auth.models import User

from offspring.web.queuemanager.models import (
    BuildRequest,
    BuildResult,
    Lexbuilder,
    ProjectGroup,
    Project,
    Release)

_marker = object()


class ObjectFactory(object):

    def __init__(self):
        self.counter = count()

    def get_unique_email_address(self):
        """Return an email address unique to this factory instance."""
        return "%s@example.com" % self.get_unique_string('email')

    def get_unique_string(self, prefix=None):
        """Return a string unique to this factory instance.

        :param prefix: Used as a prefix for the unique string. If unspecified,
            defaults to 'generic-string'.
        """
        if prefix is None:
            prefix = "generic-string"
        string = "%s%s" % (prefix, self.get_unique_integer())
        return string.lower()

    def get_unique_integer(self):
        """Return an integer unique to this factory instance."""
        return self.counter.next()

    def make_user(self):
        """Create a new User."""
        userid = password = self.get_unique_string()
        return User.objects.create_user(
            userid, self.get_unique_email_address(), password)

    def make_project(self, name=None, title=None, project_group=None):
        """
        Create a Project with the given name and title. If any of those is not
        given we use an arbitrary value.
        """
        if name is None:
            name = self.get_unique_string()
        if title is None:
            title = self.get_unique_string()
        return Project.objects.create(
            name=name, title=title, project_group=project_group)

    def make_project_group(self, name=None, title=None):
        """
        Create a ProjectGroup with the given name and title. If any of those
        is not given we use an arbitrary value.
        """
        if name is None:
            name = self.get_unique_string()
        if title is None:
            title = self.get_unique_string()
        return ProjectGroup.objects.create(name=name, title=title)

    def make_build_result(self, project=None, name=None, result=None,
                          builder=None, requestor=None, **kwargs):
        """
        Create a BuildResult with the given project or a newly created one if
        none is given.
        """
        if name is None:
            name = self.get_unique_string()
        if project is None:
            project = self.make_project()
        requested_at = kwargs.pop("requested_at", None)
        build_result = BuildResult.objects.create(
            project=project, name=name, result=result, builder=builder,
            requestor=requestor, **kwargs)
        # requested_at is auto_now_add, so we can't pass it into the object
        # creation function.
        if requested_at:
            build_result.requested_at = requested_at
            build_result.save()
        return build_result

    def make_release(self, build=None, name=None, creator=None):
        """
        Create a Release with the given build, name and creator. If any
        of those is not given we use an arbitrary value.
        """
        if build is None:
            build = self.make_build_result()
        if name is None:
            name = self.get_unique_string()
        if creator is None:
            creator = self.make_user()
        return Release.objects.create(build=build, name=name, creator=creator)

    def make_lexbuilder(self, name=None, uri=None, current_job=None):
        """
        Create a Lexbuilder with the given name, uri and current_job. If any
        of those is not given we use an arbitrary value.
        """
        if name is None:
            name = self.get_unique_string()
        if uri is None:
            uri = 'htt://example.com/%s' % name
        if current_job is _marker:
            current_job = self.make_build_result()
        return Lexbuilder.objects.create(
            name=name, uri=uri, current_job=current_job)

    def make_build_request(self, project=None):
        """
        Create a BuildRequest, if no project is given, create a project and
        associate the build with the newly created project.
        """
        if project is None:
            project = self.make_project()
        return BuildRequest.objects.create(project=project)


factory = ObjectFactory()
