# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from django.test import TestCase

from django_group_access.models import AccessGroup

from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.models import (
    ARCH_CHOICES,
    LaunchpadProject,
    ProjectGroup,
)
from offspring.web.queuemanager.forms import (
    CreateProjectForm,
    EditProjectForm,
    LaunchpadProjectForm,
    ProjectGroupForm,
)
from offspring.web.queuemanager.widgets import SelectWithAddNew


class CreateProjectFormTests(TestCase):

    def test_form_all_architecture_choices_listed(self):
        """
        Test that all architectures specified on the backend are offered as
        choices on the frontend.
        """
        form = CreateProjectForm({})
        self.assertEqual(form.fields['arch'].choices[1:], list(ARCH_CHOICES))

    def test_form_renders_only_access_groups_user_is_in(self):
        """
        A project form will only list the access groups the current user is in.
        """
        user = factory.make_user()
        group1 = AccessGroup.objects.create(name='group1')
        AccessGroup.objects.create(name='group2')
        group1.members.add(user)

        from django_group_access.middleware import _storage
        _storage.access_control_user = user

        form = CreateProjectForm({})
        self.assertEqual(form.fields['access_groups'].queryset.count(), 1)
        self.assertEqual(form.fields['access_groups'].queryset[0], group1)

        _storage.access_control_user = None

    def test_form_renders_all_access_groups_for_superuser(self):
        """
        A project form will show all access groups if the currently logged
        in user is a superuser.
        """
        user = factory.make_user()
        user.is_superuser = True
        group1 = AccessGroup.objects.create(name='group1')
        group2 = AccessGroup.objects.create(name='group2')
        group1.members.add(user)

        from django_group_access.middleware import _storage
        _storage.access_control_user = user

        form = CreateProjectForm({})
        self.assertEqual(form.fields['access_groups'].queryset.count(), 2)
        self.assertEqual(
            set(form.fields['access_groups'].queryset), set([group1, group2]))

        _storage.access_control_user = None

    def test_select_with_add_new_widget_used_for_project_group_field(self):
        """
        The project group field should use the SelectWithAddNew widget to
        allow users to create new project groups via popup.
        """
        form = CreateProjectForm()
        widget = form.fields['project_group'].widget
        self.assertTrue(
            isinstance(widget, SelectWithAddNew),
            "project_group field widget is not SelectWithAddNew.")

    def test_select_with_add_new_widget_used_for_lp_project_field(self):
        """
        The launchpad project field should use the SelectWithAddNew widget to
        allow users to create new launchpad projects via popup.
        """
        form = CreateProjectForm()
        widget = form.fields['launchpad_project'].widget
        self.assertTrue(
            isinstance(widget, SelectWithAddNew),
            "launchpad_project field widget is not SelectWithAddNew.")


class EditProjectFormTests(TestCase):

    def test_select_with_add_new_widget_used_for_project_group_field(self):
        """
        The project group field should use the SelectWithAddNew widget to
        allow users to create new project groups via popup.
        """
        form = EditProjectForm()
        self.assertTrue(
            isinstance(form.fields['project_group'].widget, SelectWithAddNew),
            "project_group field widget is not SelectWithAddNew.")

    def test_select_with_add_new_widget_used_for_lp_project_field(self):
        """
        The launchpad project field should use the SelectWithAddNew widget to
        allow users to create new launchpad projects via popup.
        """
        form = EditProjectForm()
        widget = form.fields['launchpad_project'].widget
        self.assertTrue(
            isinstance(widget, SelectWithAddNew),
            "launchpad_project field widget is not SelectWithAddNew.")


class LaunchpadProjectFormTests(TestCase):

    def test_form_is_invalid_with_no_data(self):
        """
        LaunchpadProjectForm is based on the LaunchpadProject, so requires a
        name value.
        """
        form = LaunchpadProjectForm({})
        self.assertFalse(form.is_valid())
        self.assertEqual({"name": [u"This field is required."]},
                         form.errors)

    def test_form_validates_with_all_fields(self):
        """
        Supplying a valid name to the form makes it valid.
        """
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertTrue(form.is_valid())

    def test_form_is_invalid_with_name_too_long(self):
        """
        The name for a LaunchpadProject cannot be longer than 200 characters.
        """
        form = LaunchpadProjectForm({"name": u"x"*201})
        self.assertFalse(form.is_valid())
        self.assertEqual({"name": [u"Ensure this value has at most 200 "
                                    "characters (it has 201)."]},
                         form.errors)

    def test_title_cant_be_edited_from_form(self):
        """
        The title of a LaunchpadProjectForm can't be edited.
        """
        form = LaunchpadProjectForm({"name": u"testing", "title": u"Testing"})
        self.assertTrue(form.is_valid())
        self.assertFalse("title" in form.cleaned_data,
                         "title should not be editable")

    def test_form_cleans_name_to_lower_case(self):
        """
        LaunchpadProjectForm normalises the name supplied to be all lower case.
        """
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertTrue(form.is_valid())
        self.assertEqual(u"testing", form.cleaned_data["name"])

    def test_validates_name_is_unique(self):
        """
        LaunchpadProjectForm should check that the name of the LaunchpadProject
        to be added is unique.
        """
        LaunchpadProject.objects.create(name=u"testing")
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            {"name": [u"Launchpad project with this Launchpad project name "
                       "already exists."]}, form.errors)


class ProjectGroupFormTests(TestCase):

    def test_form_cleans_name_to_lower_case(self):
        """
        ProjectGroupForm normalises the name supplied to be all lower case.
        """
        form = ProjectGroupForm({"name": u"Testing", "title": "Test Group"})
        self.assertTrue(form.is_valid())
        self.assertEqual(u"testing", form.cleaned_data["name"])
