# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for resources exposed via piston RESTful API."""

from datetime import datetime
import httplib
import simplejson
import urllib

from django.contrib.auth.models import Permission
from django.core.urlresolvers import reverse
from django.test import TestCase

from offspring.web.queuemanager.models import (
    LaunchpadProject,
    LaunchpadProjectMilestone,
    ProjectNotificationSubscription,
)
from offspring.web.queuemanager.tests.factory import factory


class BuildRequestAPITests(TestCase):
    """Tests for buildrequests piston RESTful API."""

    def setUp(self):
        self.project = factory.make_project()
        self.user = factory.make_user()
        self.buildrequest = factory.make_build_request(self.project)
        self.can_build_permission = Permission.objects.get(
            codename="can_build")
        self.project_buildrequests_url = reverse(
            'offspring.web.urls.buildrequest_handler',
            args=[self.project.name])
        self.project_buildrequests_with_id_url = reverse(
            'offspring.web.urls.buildrequest_handler',
            args=[self.project.name, self.buildrequest.id])
        self.client.login(
            username=self.user.username, password=self.user.username)

    def test_get_buildrequests_for_project(self):
        """Attempt to get all build requests for a specific project."""
        response = self.client.get(self.project_buildrequests_url)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(simplejson.loads(response.content)), 1)
    
    def test_get_buildrequest_for_project_by_id(self):
        """Attempt to get specific build request by id for a project."""
        response = self.client.get(self.project_buildrequests_with_id_url)
        self.assertEqual(response.status_code, httplib.OK)
        buildrequest = simplejson.loads(response.content)
        self.assertEqual(buildrequest['id'], self.buildrequest.id)
        self.assertEqual(buildrequest['project']['name'], self.project.name)

    def test_create_buildrequest(self):
        """Attempt to create buildrequest."""
        self.user.user_permissions.add(self.can_build_permission)
        response = self.client.post(self.project_buildrequests_url, {})
        self.assertEqual(response.status_code, httplib.CREATED)

    def test_create_buildrequest_with_reason(self):
        """Attempt to create buildrequest with a build reason comment."""
        self.user.user_permissions.add(self.can_build_permission)
        build_reason_comment = 'This is a test build reason comment.'
        response = self.client.post(
            self.project_buildrequests_url,
            { 'build_reason' : build_reason_comment,})
        self.assertEqual(response.status_code, httplib.CREATED)
        self.assertEqual(
            self.project.buildrequest_set.latest('created_at').reason,
            build_reason_comment)

    def test_create_buildrequest_without_necessary_permissions(self):
        """Attempt to create buildrequest without necessary permissions."""
        response = self.client.post(self.project_buildrequests_url, {})
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_create_buildrequest_as_anonymous_user(self):
        """Attempt to create buildrequest as an anonymous user."""
        self.client.logout()
        response = self.client.post(self.project_buildrequests_url, {})
        expected_url = urllib.quote(self.project_buildrequests_url)
        self.assertRedirects(response, '/admin/?next=%s' % expected_url)

    def test_delete_buildrequest(self):
        """Attempt to delete project buildrequests."""
        self.assertTrue(
            self.project.buildrequest_set.exists(), 
            msg="There are no buildrequests setup to try and delete.")
        self.user.user_permissions.add(self.can_build_permission)
        response = self.client.delete(self.project_buildrequests_url, {})
        self.assertEqual(response.status_code, httplib.NO_CONTENT)
        self.assertFalse(self.project.buildrequest_set.exists())

    def test_delete_buildrequest_without_necessary_permissions(self):
        """Attempt to delete buildrequests w/o necessary permissions."""
        self.assertTrue(
            self.project.buildrequest_set.exists(), 
            msg="There are no buildrequests setup to try and delete.")
        response = self.client.delete(self.project_buildrequests_url, {})
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)
        self.assertTrue(self.project.buildrequest_set.exists())


class BuildResultAPITests(TestCase):
    """Tests for buildresults piston RESTful API."""

    def setUp(self):
        self.project_with_builds = factory.make_project()
        self.builder = factory.make_lexbuilder()
        self.buildresult_a = factory.make_build_result(
            project=self.project_with_builds, builder=self.builder)
        self.buildresult_b = factory.make_build_result(
            project=self.project_with_builds, builder=self.builder)
        self.buildresult_for_random_project = factory.make_build_result()
        self.project_without_builds = factory.make_project()

    def test_get_all_buildresults(self):
        """Attempt to get all buildresults."""
        response = self.client.get(
            reverse('offspring.web.urls.buildresult_handler'))
        buildresults = simplejson.loads(response.content)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(buildresults), 3)

    def test_get_buildresults_for_project(self):
        """Attempt to get build results for project."""
        response = self.client.get(
            reverse(
                'offspring.web.urls.buildresult_handler', 
                kwargs={'projectName' : self.project_with_builds.name}
            ))
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(simplejson.loads(response.content)), 2)

    def test_get_buildresults_for_project_not_yet_built(self):
        """Attempt to get build results for project not yet built."""
        response = self.client.get(
            reverse(
                'offspring.web.urls.buildresult_handler',
                kwargs={'projectName' : self.project_without_builds.name}
            ))
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(simplejson.loads(response.content)), 0)

    def test_get_a_buildresult_for_project_by_build_name(self):
        """Attempt to get specific buildresult for a project by it's name."""
        response = self.client.get(
            reverse(
                'offspring.web.urls.buildresult_handler',
                kwargs = {
                    'projectName' : self.project_with_builds.name,
                    'buildName' : self.buildresult_a.name,
                },
            ))
        buildresults = simplejson.loads(response.content)
        buildresult = buildresults[0]
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(buildresults), 1)
        self.assertEqual(buildresult['name'], self.buildresult_a.name)
        self.assertEqual(
            buildresult['project']['name'], self.buildresult_a.project.name)

    def test_get_buildresult_for_builder(self):
        """Attempt to get all buildresults produced by a specific builder."""
        response = self.client.get(
            reverse(
                'offspring.web.urls.buildresult_handler',
                kwargs = {
                    'builderName' : self.builder.name,
                },
            ))
        buildresults = simplejson.loads(response.content)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(buildresults), 2)


class LaunchpadProjectMilestoneAPITests(TestCase):
    """Tests for launchpad project milestones piston RESTful API."""

    def setUp(self):
        self.lp_project = LaunchpadProject.objects.create(name='test-project')
        self.lp_milestone = LaunchpadProjectMilestone.objects.create(
            name='test-milestone',
            launchpad_project=self.lp_project,
            date_targeted=datetime.today())
        self.view_url = reverse('offspring.web.urls.milestone_handler')

    def test_get_launchpad_project_milestones(self):
        """Attempt to get all launchapd project milestones."""
        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(simplejson.loads(response.content)), 1)


class ProjectNotificationSubscriptionAPITests(TestCase):
    """Tests for project notificaiton subscriptions piston RESTful API."""

    def setUp(self):
        self.project = factory.make_project()
        self.user = factory.make_user()
        self.view_url = reverse(
            'offspring.web.urls.projectNotificationSubscription_handler',
            args=[self.project.name])
        self.client.login(
            username=self.user.username, password=self.user.username)

    def test_get_project_notification_subscription(self):
        """Attempt to get project notification subscription for project."""
        subscription = ProjectNotificationSubscription.objects.create(
            user=self.user, project=self.project)
        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(
            subscription.id, simplejson.loads(response.content)['id'])

    def test_get_nonexistant_project_notification_subscription(self):
        """Attempt to get subscription for nonexistent project."""
        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_get_project_notification_subscription_as_anonymous_user(self):
        """Attempt to get notification subscription as anonymous user."""
        self.client.logout()
        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_create_project_notification_subscription(self):
        """Attempt to create project notification subscription."""
        response = self.client.post(self.view_url, {})
        self.assertEqual(response.status_code, httplib.OK)
        subscription = ProjectNotificationSubscription.objects.get(
            id=simplejson.loads(response.content)['id'])
        self.assertEqual(self.project.name, subscription.project.name)

    def test_create_notification_subscription_for_nonexistent_project(self):
        """Attempt to create subscription for nonexistent project."""
        self.view_url = reverse(
            'offspring.web.urls.projectNotificationSubscription_handler',
            args=['nonexistent_project'])
        response = self.client.post(self.view_url, {})
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_create_duplicate_project_notification_subscription(self):
        """Attempt to create duplicate subscription for project."""
        ProjectNotificationSubscription.objects.create(
            user=self.user, project=self.project)
        response = self.client.post(self.view_url, {})
        self.assertEqual(response.status_code, httplib.CONFLICT)

    def test_delete_project_notification_subscription(self):
        """Attempt to delete project notification subscription."""
        subscription = ProjectNotificationSubscription.objects.create(
            user=self.user, project=self.project)
        response = self.client.delete(self.view_url)
        self.assertEqual(response.status_code, httplib.NO_CONTENT)
        self.assertRaises(
            ProjectNotificationSubscription.DoesNotExist,
            ProjectNotificationSubscription.objects.get, id=subscription.id)

    def test_delete_notification_subscription_for_nonexistent_project(self):
        """Attempt to delete subscription for nonexistent project."""
        response = self.client.delete(
            reverse(
                'offspring.web.urls.projectNotificationSubscription_handler',
                args=['nonexistent-project'])
            )
        self.assertEqual(response.status_code, httplib.NOT_FOUND)


class ReleaseAPITests(TestCase):
    """Tests for releases piston RESTful API."""

    def setUp(self):
        self.project = factory.make_project()
        self.build_result = factory.make_build_result(project=self.project)
        self.release = factory.make_release(build=self.build_result)

    def test_get_releases_for_project(self):
        """Attempt to get all releases belonging to a project."""
        response = self.client.get(
            reverse('offspring.web.urls.release_handler',
            args=[self.project.name]))
        releases = simplejson.loads(response.content)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(len(releases), 1)
