# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime, timedelta

from django.test import TestCase

from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.models import BuildResult

from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import (
    create_build_results_with_durations, create_build_results_with_queue_times)


class BuildResultMetricsManagerTest(TestCase):

    def setUp(self):
        self.project = factory.make_project()

    def test_get_average_build_time_single_build(self):
        """
        get_average_build_time should return the average time taken to build a
        project.
        """
        finished_at = datetime.utcnow()
        factory.make_build_result(
            project=self.project, result=ProjectBuildStates.SUCCESS,
            finished_at=finished_at,
            started_at=finished_at - timedelta(minutes=10))
        self.assertEqual(10 * 60,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_only_includes_completed_builds(self):
        """
        get_average_build_time should only include builds that are in a
        completed state.
        """
        create_build_results_with_durations(
            self.project, [10, 20], result=ProjectBuildStates.UNKNOWN)
        create_build_results_with_durations(
            self.project, [10, 20], result=ProjectBuildStates.FAILED)
        create_build_results_with_durations(
            self.project, [10, 20], result=ProjectBuildStates.PENDING)

        self.assertEqual(None,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_no_builds(self):
        """
        If there are no builds, we should get an average of None.
        """
        self.assertEqual(None,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_unfinished_build(self):
        """
        If a build has started, but not yet finished, then we should get None,
        for the average build time.
        """
        started_at = datetime.utcnow()
        factory.make_build_result(project=self.project,
                                                 started_at=started_at)
        self.assertEqual(None,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_multiple_builds(self):
        """
        The time for each BuildResult should be averaged together.
        """
        average = create_build_results_with_durations(self.project, [10, 20])
        self.assertEqual(average * 60,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_multiple_builds_ignores_unfinished(self):
        """
        Any unfinished BuildResults should be discounted from the calculation.
        """
        average = create_build_results_with_durations(self.project, [10, 20])

        # Create an additional unfinished build.
        started_at = datetime.utcnow() - timedelta(minutes=20)
        factory.make_build_result(project=self.project,
                                                 started_at=started_at)

        self.assertEqual(average * 60,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_ignores_other_projects(self):
        """
        BuildResults associated with other projects should not be included in
        the calculation.
        """
        average = create_build_results_with_durations(self.project, [10, 20])
        create_build_results_with_durations(factory.make_project(), [50, 100])
        self.assertEqual(average * 60,
                         BuildResult.metrics.get_average_build_time(
                             self.project))

    def test_get_average_build_time_no_project(self):
        """
        If get_average_build_time is not supplied with a Project, then it
        should calculate the average across all projects.
        """
        average1 = create_build_results_with_durations(self.project, [10, 20])
        average2 = create_build_results_with_durations(
            factory.make_project(), [50, 100])
        # 10 + 20 + 50 + 100 = 180 / 4 = 45 minutes...
        self.assertEqual((average1 + average2) / 2.0 * 60,
                         BuildResult.metrics.get_average_build_time())

    def test_get_average_build_time_of_automated_builds_only(self):
        """
        Calling get_average_build_time with automated=True should filter the
        results to only include automated builds, which are defined as builds
        without a requestor.
        """
        average = create_build_results_with_durations(self.project, [10, 20])
        self.assertEqual(
            average * 60.0,
            BuildResult.metrics.get_average_build_time(automated=True))

    def test_get_average_build_time_with_non_automated_builds(self):
        """
        Calling get_average_build_time with automated=False should filter the
        results to only include manual builds, which are defined as builds
        with a requestor.
        """
        user = factory.make_user()
        average = create_build_results_with_durations(
            factory.make_project(), [50, 100], requestor=user)
        self.assertEqual(
            average * 60.0,
            BuildResult.metrics.get_average_build_time(automated=False))

    def test_get_average_build_time_filters_by_date(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame.

        Any BuildResults finishing outwith the time period specified shouldn't
        be included.
        """
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_durations(
            self.project, [10, 20], build_date=build_date)

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            None,
            BuildResult.metrics.get_average_build_time(period=period))

    def test_get_average_build_time_includes_by_date(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame, specifying a period should only include those BuildResults
        within that period.
        """
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_durations(
            self.project, [10, 20], build_date=build_date)

        average = create_build_results_with_durations(
            factory.make_project(), [50, 100],
            build_date=build_date + timedelta(days=2))

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_build_time(period=period))

    def test_get_average_build_time_filters_by_date_and_automated(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame, specifying a period and the automated flag should filter the
        applicable BuildResults.
        """
        user = factory.make_user()
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_durations(
            self.project, [10, 20], build_date=build_date,
            requestor=user)
        average = create_build_results_with_durations(
            factory.make_project(), [40, 40],
            build_date=build_date + timedelta(days=2))

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_build_time(period=period,
                                                       automated=True))

    def test_average_queue_time_single_build(self):
        """
        get_average_queue_time should return the average time a build waits
        before being dispatched.
        """
        dispatched_at = datetime.utcnow()
        factory.make_build_result(
            project=self.project, dispatched_at=dispatched_at,
            requested_at=dispatched_at - timedelta(minutes=30))
        self.assertEqual(
            30 * 60,
            BuildResult.metrics.get_average_queue_time(self.project))

    def test_average_queue_time_no_builds(self):
        """
        If there are no builds, we should get an average of None.
        """
        self.assertEqual(
            None,
            BuildResult.metrics.get_average_queue_time(self.project))

    def test_average_queue_time_includes_by_date(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame, specifying a period should only include those BuildResults
        within that period.
        """
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_queue_times(
            self.project, [10, 20], build_date=build_date)

        average = create_build_results_with_queue_times(
            factory.make_project(), [50, 100],
            build_date=build_date + timedelta(days=2))

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_queue_time(period=period))

    def test_get_average_queue_time_multiple_builds(self):
        """
        The time for each BuildResult should be averaged together.
        """
        create_build_results_with_queue_times(self.project, [10, 20])
        self.assertEqual(15 * 60,
                         BuildResult.metrics.get_average_queue_time(
                             self.project))

    def test_get_average_queue_time_ignores_other_projects(self):
        """
        BuildResults associated with other projects should not be included in
        the calculation.
        """
        average = create_build_results_with_queue_times(self.project, [10, 20])
        create_build_results_with_queue_times(factory.make_project(),
                                              [50, 100])
        self.assertEqual(average * 60,
                         BuildResult.metrics.get_average_queue_time(
                             self.project))

    def test_get_average_queue_time_no_project(self):
        """
        If get_average_queue_time is not supplied with a Project, then it
        should calculate the average across all projects.
        """
        average1 = create_build_results_with_queue_times(self.project, [10, 20])
        average2 = create_build_results_with_queue_times(
            factory.make_project(), [50, 100])
        # 10 + 20 + 50 + 100 = 180 / 4 = 45 minutes...
        self.assertEqual((average1 + average2) / 2.0 * 60,
                         BuildResult.metrics.get_average_queue_time())

    def test_get_average_queue_time_with_automated_builds(self):
        """
        Calling get_average_queue_time with automated=True should filter the
        results to only include automated builds, which are defined as builds
        without a requestor.
        """
        average = create_build_results_with_queue_times(self.project, [10, 20])
        self.assertEqual(
            average * 60.0,
            BuildResult.metrics.get_average_queue_time(automated=True))

    def test_get_average_queue_time_with_non_automated_builds(self):
        """
        Calling get_average_queue_time with automated=False should filter the
        results to only include manual builds, which are defined as builds
        with a requestor.
        """
        average = create_build_results_with_queue_times(self.project, [10, 20])
        self.assertEqual(
            average * 60.0,
            BuildResult.metrics.get_average_queue_time(automated=True))

    def test_get_average_queue_time_filters_by_date(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame.

        Any BuildResults finishing outwith the time period specified shouldn't
        be included.
        """
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_durations(
            self.project, [10, 20], build_date=build_date)

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            None,
            BuildResult.metrics.get_average_queue_time(period=period))

    def test_get_average_queue_time_includes_by_date(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame, specifying a period should only include those BuildResults
        within that period.
        """
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_queue_times(
            self.project, [10, 20], build_date=build_date)

        average = create_build_results_with_queue_times(
            factory.make_project(), [50, 100],
            build_date=build_date + timedelta(days=2))

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_queue_time(period=period))

    def test_get_average_queue_time_filters_by_date_and_automated(self):
        """
        It should be possible to restrict the date queried to a specific time
        frame, specifying a period and the automated flag should filter the
        applicable BuildResults.
        """
        user = factory.make_user()
        build_date = datetime.utcnow() - timedelta(days=30)
        create_build_results_with_queue_times(
            self.project, [10, 20], build_date=build_date,
            requestor=user)
        average = create_build_results_with_queue_times(
            factory.make_project(), [40, 40],
            build_date=build_date + timedelta(days=2))

        # From the day after we create our builds 'til today.
        period = (build_date + timedelta(days=1), datetime.utcnow())
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_queue_time(period=period,
                                                       automated=True))


class ProjectGroupBuildResultMetricsManagerTest(TestCase):

    def setUp(self):
        self.project_group = factory.make_project_group()
        self.project = factory.make_project(project_group=self.project_group)

    def test_get_average_build_time_with_project_group(self):
        """
        get_average_build_time should take a project group, and calculate the
        average build time of all projects within that group.
        """
        finished_at = datetime.utcnow()
        factory.make_build_result(
            project=self.project, result=ProjectBuildStates.SUCCESS,
            finished_at=finished_at,
            started_at=finished_at - timedelta(minutes=10))
        self.assertEqual(
            10 * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_build_time_with_multiple_projects_in_group(self):
        """
        All Projects in a ProjectGroup should be included in the average
        buildtime.
        """
        average1 = create_build_results_with_durations(self.project, [10, 20])
        project2 = factory.make_project(project_group=self.project_group)
        average2 = create_build_results_with_durations(project2, [20, 30])
        # 10 + 20 + 20 + 30 / 4 = 20 minutes
        self.assertEqual(
            (average1 + average2) / 2.0 * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_build_time_with_excludes_projects_not_in_group(self):
        """
        All Projects in a ProjectGroup should be included in the average
        buildtime.
        """
        average = create_build_results_with_durations(self.project, [10, 20])
        project2 = factory.make_project()
        create_build_results_with_durations(project2, [20, 30])
        # 10 + 20  / 2 = 15 minutes
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_queue_time_with_project_group(self):
        """
        get_average_queue_time should take a project group, and calculate the
        average queue time of all projects within that group.
        """
        average = create_build_results_with_queue_times(self.project, [10, 20])
        self.assertEqual(
            average * 60,
            BuildResult.metrics.get_average_queue_time(
                project_group=self.project_group))
