# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime, timedelta

from django.test import TestCase

from offspring.web.queuemanager.metrics import (
    get_build_result_metrics)

from offspring.web.queuemanager.models import BuildResult
from offspring.enums import ProjectBuildStates

from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import (
    create_build_results_with_durations, create_build_results_with_queue_times)


class GetBuildResultsTests(TestCase):

    def setUp(self):
        self.project = factory.make_project()

    def test_get_build_result_metrics(self):
        """
        get_build_result_metrics should return a list of tuples with the metric
        being measured, and the time in minutes.
        """
        def create_test_data(user=None, offset=0):
            start_date = datetime.utcnow()
            last_week = start_date - timedelta(days=8)
            last_month = start_date - timedelta(days=35)
            # Two this week, average = 10 + 20/2 = 15
            create_build_results_with_durations(
                self.project, [10 + offset, 20],
                build_date=start_date - timedelta(days=1),
                requestor=user)

            # Two this week, average = 30 + 18/2 = 24.0
            create_build_results_with_queue_times(
                self.project, [30 + offset, 18],
                build_date=start_date - timedelta(days=1),
                requestor=user)

            # Two last week, 10 + 10 + 10 + 20/4 = 12.5
            create_build_results_with_durations(
                self.project, [10 + offset, 10], build_date=last_week,
                requestor=user)

            # Two last week, average = 30 + 18 + 14  11/4 = 73
            create_build_results_with_queue_times(
                self.project, [14 + offset, 11], build_date=last_week,
                requestor=user)

            # Two last month, 10 + 10 + 10 + 20 + 30 + 40/6 = 20
            create_build_results_with_durations(
                self.project, [30 + offset, 40], build_date=last_month,
                requestor=user)

            # Two last month, average = 30 + 18 + 14  11 + 30 + 20/6 = 82
            create_build_results_with_queue_times(
                self.project, [30 + offset, 20], build_date=last_month,
                requestor=user)

        create_test_data()
        user = factory.make_user()
        create_test_data(user=user, offset=10)

        self.assertEqual(
            [("Average scheduled build time (7 days)", 15.0 * 60),
             ("Average scheduled build time (30 days)", 12.5 * 60),
             ("Average scheduled build time (90 days)", 20.0 * 60),
             ("Average scheduled queue time (7 days)", 24.0 * 60),
             ("Average scheduled queue time (30 days)", 18.25 * 60),
             ("Average scheduled queue time (90 days)", 20.5 * 60),
             ("Average requested build time (7 days)", 20.0 * 60),
             ("Average requested build time (30 days)", 17.5 * 60),
             ("Average requested build time (90 days)", 25.0 * 60),
             ("Average requested queue time (7 days)", 29.0 * 60),
             ("Average requested queue time (30 days)", 23.25 * 60),
             ("Average requested queue time (90 days)", 25.5 * 60)],
            get_build_result_metrics())


class ProjectGroupBuildMetricsTests(TestCase):

    def setUp(self):
        self.project_group = factory.make_project_group()
        self.project = factory.make_project(project_group=self.project_group)

    def test_get_average_build_time_with_project_group(self):
        """
        get_average_build_time should take a project group, and calculate the
        average build time of all projects within that group.
        """
        build_result = factory.make_build_result(
            project=self.project, result=ProjectBuildStates.SUCCESS)
        finished_at = datetime.now()
        build_result.finished_at = finished_at
        build_result.started_at = finished_at - timedelta(minutes=10)
        build_result.save()
        self.assertEqual(
            10 * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_build_time_with_multiple_projects_in_group(self):
        """
        All Projects in a ProjectGroup should be included in the average
        buildtime.
        """
        create_build_results_with_durations(self.project, [10, 20])
        project2 = factory.make_project(project_group=self.project_group)
        create_build_results_with_durations(project2, [20, 30])
        # 10 + 20 + 20 + 30 / 4 = 20 minutes
        self.assertEqual(
            20 * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_build_time_with_excludes_projects_not_in_group(self):
        """
        Only Projects in a ProjectGroup should be included in the average
        buildtime.
        """
        create_build_results_with_durations(self.project, [10, 20])
        project2 = factory.make_project()
        create_build_results_with_durations(project2, [20, 30])
        # 10 + 20  / 2 = 15 minutes
        self.assertEqual(
            15 * 60,
            BuildResult.metrics.get_average_build_time(
                project_group=self.project_group))

    def test_get_average_queue_time_with_project_group(self):
        """
        get_average_queue_time should take a project group, and calculate the
        average queue time of all projects within that group.
        """
        create_build_results_with_queue_times(self.project, [10, 20])
        self.assertEqual(
            15 * 60,
            BuildResult.metrics.get_average_queue_time(
                project_group=self.project_group))
