# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime, timedelta, date

from django.test import TestCase
from mocker import MockerTestCase

from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import (
    create_build_results_with_durations
)

class ProjectTests(TestCase):

    def test_average_build_time(self):
        """
        Project.average_build_time should use the BuildResult metric.
        """
        project = factory.make_project()
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS)
        self.assertEqual(50 * 60.0, project.average_build_time)


class LexbuilderTests(TestCase, MockerTestCase):

    def test_estimated_completion(self):
        """
        Lexbuilder.estimated_completion_time should be calculated as:
            average build time - (current time - start time)
        """
        project = factory.make_project()

        create_build_results_with_durations(
            project=project, durations=[25, 25],
            result=ProjectBuildStates.SUCCESS)

        started_at = datetime.now() - timedelta(minutes=10)
        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING,
            started_at=started_at)
        builder = factory.make_lexbuilder(current_job=build_result)

        datetime_mock = self.mocker.replace(datetime)
        datetime_mock.now()
        self.mocker.result(started_at + timedelta(minutes=5))
        self.mocker.replay()

        # Average time = 25 minutes
        # Started 5 minutes ago (started_at + timedelta(minutes=5))
        # 20 minutes remaining

        self.assertEqual(20 * 60, builder.estimated_completion)

    def test_estimated_completion_with_no_average_time(self):
        """
        If the average cannot be determined because there are no builds, we
        should get "Unknown" as the estimated completion time.
        """
        project = factory.make_project()

        started_at = datetime.now() - timedelta(minutes=10)
        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING,
            started_at=started_at)

        builder = factory.make_lexbuilder(current_job=build_result)

        self.assertEqual("Unknown", builder.estimated_completion)

    def test_estimated_completion_with_no_current_build(self):
        """
        If there is no current build, then we should get "No build in
        progress" for the estimated completion time.
        """
        builder = factory.make_lexbuilder(current_job=None)
        self.assertEqual("No build in progress", builder.estimated_completion)

    def test_estimated_completion_negative_time(self):
        """
        If the current elapsed time is higher than the average, we should get
        "ASAP" as the estimated completion.
        """
        project = factory.make_project()

        create_build_results_with_durations(
            project=project, durations=[5, 5],
            result=ProjectBuildStates.SUCCESS)

        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING)
        started_at = datetime.now() - timedelta(minutes=10)
        build_result.started_at = started_at
        build_result.save()

        builder = factory.make_lexbuilder(current_job=build_result)

        datetime_mock = self.mocker.replace(datetime)
        datetime_mock.now()
        self.mocker.result(started_at + timedelta(minutes=10))
        self.mocker.replay()

        # Average time = 5 minutes
        # Started 10 minutes ago (started_at + timedelta(minutes=10))
        self.assertEqual("ASAP", builder.estimated_completion)

    def test_average_load(self):
        """
        Lexbuilder.average_load should calculate the total time used by builds
        as a percentage of the time available.
        """
        builder = factory.make_lexbuilder()
        project = factory.make_project()
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS,
            build_date=datetime.now() - timedelta(days=2),
            builder=builder)
        # 50 minutes * 60 seconds + 50 minutes * 60 seconds / (24 * 60 * 60)
        # = 0.0694 = 6.94%
        self.assertEqual(6.94, round(builder.average_load, 2))

    def test_average_load_buildresults_today(self):
        """
        Lexbuilder.average_load should ignore buildresults from today.
        """
        builder = factory.make_lexbuilder()
        project = factory.make_project()

        create_build_results_with_durations(
            project, [50, 50], build_date=datetime.now().date(),
            result=ProjectBuildStates.SUCCESS, builder=builder)
        self.assertEqual(None, builder.average_load)


class ProjectGroupTests(TestCase):

    def test_average_build_time(self):
        """
        ProjectGroup.average_build_time should use the BuildResult metric.
        """
        project_group = factory.make_project_group()
        project = factory.make_project(project_group=project_group)
        create_build_results_with_durations(
            project, [10, 20], result=ProjectBuildStates.SUCCESS)
        self.assertEqual(15 * 60.0, project_group.average_build_time)
