# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import urllib

from django.contrib.auth.models import Permission
from django.core.urlresolvers import reverse
from django.test import TestCase

from django_group_access.models import AccessGroup

from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.models import (
    LaunchpadProject,
    ProjectGroup,
)
from offspring.web.queuemanager.forms import (
    LaunchpadProjectForm,
    ProjectGroupForm,
)
from offspring.web.queuemanager.tests.factory import factory


class ProjectDetailsViewTests(TestCase):

    def setUp(self):
        self.project = factory.make_project()
        self.view_url = reverse(
            'offspring.web.queuemanager.views.project_details',
            args=[self.project.name])

    def test_project_details_view(self):
        """
        Test project details view.
        """
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'Project Details for %s' % self.project.title.capitalize(),
            status_code=200,
            msg_prefix=response.content
        )


class AccessTests(TestCase):
    """
    Tests for record visibility.
    """
    def setUp(self):
        self.project1 = factory.make_project()
        self.project2 = factory.make_project()
        self.group, _ = AccessGroup.objects.get_or_create(name='Group')
        self.project2.access_groups.add(self.group)
        self.projects_url = reverse(
            'offspring.web.queuemanager.views.projects')
        self.builds_url = reverse(
            'offspring.web.queuemanager.views.builds')

    def test_project_visibility_anonymous_user(self):
        """
        Only unrestricted projects (with no access groups
        specified) are visibile to anonymous users.
        """
        response = self.client.get(self.projects_url)
        self.assertContains(response, self.project1.title)
        self.assertNotContains(response, self.project2.title)

    def test_project_visibility_authenticated(self):
        """
        Restricted projects are visibile to users in the correct
        group.
        """
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.projects_url)
        self.assertContains(response, self.project1.title)
        self.assertContains(response, self.project2.title)

    def test_build_result_visibility_anonymous_user(self):
        """
        Only builds for unrestricted projects are visible to
        anonymous users.
        """
        result = factory.make_build_result(self.project1)
        hidden_result = factory.make_build_result(self.project2)
        response = self.client.get(self.builds_url)
        self.assertContains(response, result.name)
        self.assertNotContains(response, hidden_result.name)

    def test_build_result_visibility_authenticated_user(self):
        """
        Builds for restricted projects are visibile to users in the correct
        group.
        """
        result = factory.make_build_result(self.project1)
        hidden_result = factory.make_build_result(self.project2)
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.builds_url)
        self.assertContains(response, result.name)
        self.assertContains(response, hidden_result.name)

    def test_build_request_visibility_anonymous_user(self):
        """
        Only build requests for unrestricted projects are visible to
        anonymous users.
        """
        request = factory.make_build_request(self.project1)
        hidden_request = factory.make_build_request(self.project2)
        response = self.client.get(self.projects_url)
        self.assertContains(response, str(request))
        self.assertNotContains(response, str(hidden_request))

    def test_build_request_visibility_authenticated_user(self):
        """
        Build requests for restricted projects are visibile to users in the
        correct group.
        """
        request = factory.make_build_request(self.project1)
        hidden_request = factory.make_build_request(self.project2)
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.projects_url)
        self.assertContains(response, str(request))
        self.assertContains(response, str(hidden_request))


class ProjectEditViewTests(TestCase):

    def setUp(self):
        self.project = factory.make_project()
        self.view_url = reverse(
            'offspring.web.queuemanager.views.project_edit',
            args=[self.project.name])

    def test_project_edit_view(self):
        """
        Test project edit view with authenticated user.
        """
        p = Permission.objects.get(codename="change_project")
        user = factory.make_user()
        user.user_permissions.add(p)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'Update details for %s' % self.project.title.capitalize(),
            status_code=200,
            msg_prefix=response.content
        )


class BuildResultViewTests(TestCase):

    def test_buildresult_details_view(self):
        """
        Test BuildResult details view.
        """
        project = factory.make_project()
        build = factory.make_build_result(project=project)
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_details',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Build Details', status_code=200,
            msg_prefix=response.content)


class BuildResultPublishViewTests(TestCase):

    def test_buildresult_publish_view(self):
        """
        Test build result publish view with authorized user.
        """
        project = factory.make_project()
        build = factory.make_build_result(
            project=project, result=ProjectBuildStates.SUCCESS)
        user = factory.make_user()
        p = Permission.objects.get(codename="add_release")
        user.user_permissions.add(p)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_publish',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Release %s build' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)


class ProjectBuildsViewTests(TestCase):

    def test_project_builds_view(self):
        """
        Test project build result list view.
        """
        project = factory.make_project()
        factory.make_build_result(project=project)
        response = self.client.get(
            reverse(
                'offspring.web.queuemanager.views.builds',
                args=[project.name],
            ),
        )
        self.assertContains(
            response, 'Build Records for %s' % project.title,
            status_code=200, msg_prefix=response.content)


class OverallBuildsViewTests(TestCase):

    def test_builds_view(self):
        """
        Test build result list view.
        """
        project = factory.make_project()
        build = factory.make_build_result(project=project)
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.builds'))
        self.assertContains(
            response, build.name, status_code=200,
            msg_prefix=response.content)


class LaunchpadProjectViewsTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.launchpad_project_create")
        self.user = factory.make_user()

    def test_launchpad_project_create_view_redirects_unauthorized_user(self):
        """
        If the logged in user doesn't have permission to access the form, they
        should be directed to login.
        """
        self.client.login(username=self.user.username,
                          password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertRedirects(
            response,
            "/admin/?next=" + urllib.quote(self.view_url),
        )

    def test_get_launchpad_project_create_view(self):
        """
        If the logged in user has the correct permissions, a GET request
        should get the LaunchpadProjectForm.
        """
        p = Permission.objects.get(codename="add_launchpadproject")
        self.user.user_permissions.add(p)
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.view_url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], LaunchpadProjectForm),
            "Form is not LaunchpadProjectForm")
        self.assertTemplateUsed(
            response, "queuemanager/launchpad_project_create_popup.html")

    def test_post_launchpad_project_create_view(self):
        """
        A POST request should create a new LaunchpadProject with the entered
        name.
        """
        p = Permission.objects.get(codename="add_launchpadproject")
        self.user.user_permissions.add(p)
        self.client.login(username=self.user.username,
                          password=self.user.username)
        self.assertQuerysetEqual(LaunchpadProject.objects.all(), [])
        response = self.client.post(
            self.view_url, {"name": u"test"})
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(
            LaunchpadProject.objects.all(), ['<LaunchpadProject: test>'])


class ProjectGroupFormTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.projectgroup_create")
        self.user = factory.make_user()

    def test_projectgroup_create_view_redirects_unauthorized_user(self):
        """
        If the logged in user doesn't have permission to access the form, they
        should be directed to login.
        """
        self.client.login(username=self.user.username,
                          password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertRedirects(
            response,
            "/admin/?next=" + urllib.quote(self.view_url),
        )

    def test_get_projectgroup_create_view(self):
        """
        If the logged in user has the correct permissions, a GET request
        should get the ProjectGroupForm.
        """
        add_projectgroup_permission = Permission.objects.get(
            codename="add_projectgroup")
        self.user.user_permissions.add(add_projectgroup_permission)
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.view_url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], ProjectGroupForm),
            "Form is not ProjectGroupForm")
        self.assertTemplateUsed(
            response, "queuemanager/projectgroup_create_popup.html")

    def test_post_projectgroup_create_view(self):
        """
        A POST request should create a new ProjectGroup with the entered
        name and title.
        """
        add_projectgroup_permission = Permission.objects.get(
            codename="add_projectgroup")
        self.user.user_permissions.add(add_projectgroup_permission)
        self.client.login(username=self.user.username,
                          password=self.user.username)
        post_data = {"name": u"test", "title": u"Test ProjectGroup"}
        response = self.client.post(self.view_url, post_data)
        self.assertContains(response, text='opener.dismissAddAnotherPopup')
        self.assertEqual(ProjectGroup.objects.values()[0], post_data)
